#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    The edc_noise_handling module provides various methods for noise
    compensation of room impulse responses.
"""

import numpy as np
from matplotlib import pyplot as plt
from pyrato import dsp
import pyrato as ra
import warnings


def estimate_noise_energy(
        data,
        interval=[0.9, 1.0],
        is_energy=False):
    """ This function estimates the noise energy level of a given room impulse
    response. The noise is assumed to be Gaussian.

    Parameters
    ----------
    data: np.array
        The room impulse response with dimension [..., n_samples]
    interval : tuple, float, [0.9, 1.]
        Defines the interval of the RIR to be evaluated for the estimation.
        The interval is relative to the length of the RIR [0 = 0%, 1=100%)]
    is_energy: Boolean
        Defines if the data is already squared.

    Returns
    -------
    noise_energy: float
        The energy of the background noise.
    """

    energy_data = preprocess_rir(
        data,
        is_energy=is_energy,
        time_shift=False,
        channel_independent=False)[0]

    if np.any(energy_data) < 0:
        raise ValueError("Energy is negative, check your input signal.")

    region_start_idx = int(energy_data.shape[-1]*interval[0])
    region_end_idx = int(energy_data.shape[-1]*interval[1])
    mask = np.arange(region_start_idx, region_end_idx)
    noise_energy = np.nanmean(np.take(energy_data, mask, axis=-1), axis=-1)

    return noise_energy


def preprocess_rir(
        data,
        is_energy=False,
        time_shift=False,
        channel_independent=False):
    """ Preprocess the room impulse response for further processing:
        - Square data
        - Shift the RIR to the first sample of the array, compensating for the
          delay of the time of arrival of the direct sound. The time shift is
          performed as a non-cyclic shift, adding numpy.nan values in the end
          of the RIR corresponding to the number of samples the data is
          shifted by.
        - The time shift can be done channel-independent or not.

    Parameters
    ----------
    data : ndarray, double
        The room impulse response with dimension [..., n_samples]
    is_energy : boolean
        Defines, if the data is already squared.
    time_shift : boolean
        Defines, if the silence at beginning of the RIR should be removed.
    channel_independent : boolean
        Defines, if the time shift is done channel-independent or not.

    Returns
    -------
    energy_data : ndarray, double
        The preprocessed RIR
    n_channels : integer
        The number of channels of the RIR
    data_shape : list, integer
        The original data shape.

    """
    n_samples = data.shape[-1]
    data = np.atleast_2d(data)
    n_channels = np.prod(data.shape[:-1])

    data_shape = list(data.shape)
    data = np.reshape(data, (-1, n_samples))

    if time_shift:
        rir_start_idx = dsp.find_impulse_response_start(data)

        if channel_independent and not n_channels == 1:
            shift_samples = -rir_start_idx
        else:
            min_shift = np.amin(rir_start_idx)
            shift_samples = np.asarray(
                -min_shift * np.ones(n_channels), dtype=int)

        result = dsp.time_shift(
            data, shift_samples, circular_shift=False, keepdims=True)

    else:
        result = data
    if not is_energy:
        energy_data = np.abs(result)**2
    else:
        energy_data = result.copy()

    return energy_data, n_channels, data_shape


def smooth_rir(
        data,
        sampling_rate,
        smooth_block_length=0.075):
    """ Smoothens the RIR by averaging the data in an specified interval.

    Parameters
    ----------
    data : ndarray, double
        The room impulse response with dimension [..., n_samples]
    sampling_rate: integer
        Defines the sampling rate of the room impulse response.
    smooth_block_length : double
        Defines the block-length of the smoothing algorithm in seconds.

    Returns
    -------
    time_window_data : ndarray, double
        The smoothed RIR.
    time_vector_window : ndarray, double
        The respective time vector fitting the smoothed data.
    time_vector : ndarray, double
        The time vector fitting the original data.

    """
    data = np.atleast_2d(data)
    n_samples = data.shape[-1]
    n_samples_nan = np.count_nonzero(np.isnan(data), axis=-1)

    n_samples_per_block = int(np.round(smooth_block_length * sampling_rate, 0))
    n_blocks = np.asarray(
        np.floor((n_samples-n_samples_nan)/n_samples_per_block),
        dtype=int)

    n_blocks_min = int(np.min(n_blocks))
    n_samples_actual = int(n_blocks_min*n_samples_per_block)
    reshaped_array = np.reshape(
        data[..., :n_samples_actual],
        (-1, n_blocks_min, n_samples_per_block))
    time_window_data = np.mean(reshaped_array, axis=-1)

    # Use average time instances corresponding to the average energy level
    # instead of time for the first sample of the block
    time_vector_window = \
        ((0.5+np.arange(0, n_blocks_min)) * n_samples_per_block/sampling_rate)

    # Use the time corresponding to the sampling of the original data
    time_vector = (np.arange(0, n_samples))/sampling_rate

    return time_window_data, time_vector_window, time_vector


def subtract_noise_from_squared_rir(data, noise_level='auto'):
    """ Subtracts the noise energy level from the squared RIR. Note, that the
        RIR has to be squared before.

    Parameters
    ----------
    data : ndarray, double
        The squared room impulse response with dimension [..., n_samples]

    Returns
    -------
    result : ndarray, double
        The noise-subtracted RIR.

    """
    if noise_level == "auto":
        noise_level = estimate_noise_energy(
            data,
            is_energy=True,
            interval=[0.9, 1.0])
    return (data.T - noise_level).T


def energy_decay_curve_truncation(
        data,
        sampling_rate,
        freq='broadband',
        noise_level='auto',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False):
    """ This function truncates a given room impulse response by the
    intersection time after Lundeby and calculates the energy decay curve.

    Parameters
    ----------
    data : ndarray, double
        The room impulse response with dimension [..., n_samples]
    sampling_rate: integer
        The sampling rate of the room impulse response.
    freq: integer OR string
        The frequency band. If set to 'broadband',
        the time window of the Lundeby-algorithm will not be set in dependence
        of frequency.
    noise_level: ndarray, double OR string
        If not specified, the noise level is calculated based on the last 10
        percent of the RIR. Otherwise specify manually for each channel
        as array.
    is_energy: boolean
        Defines, if the data is already squared.
    time_shift : boolean
        Defines, if the silence at beginning of the RIR should be removed.
    channel_independent : boolean
        Defines, if the time shift and normalization is done
        channel-independently or not.
    normalize : boolean
        Defines, if the energy decay curve should be normalized in the end
        or not.
    plot: Boolean
        Specifies, whether the results should be visualized or not.

    Returns
    -------
    energy_decay_curve: ndarray, double
        Returns the noise handeled edc.
    """
    energy_data, n_channels, data_shape = preprocess_rir(
        data,
        is_energy=is_energy,
        time_shift=time_shift,
        channel_independent=channel_independent)
    n_samples = energy_data.shape[-1]

    intersection_time = intersection_time_lundeby(
        energy_data,
        sampling_rate=sampling_rate,
        freq=freq,
        initial_noise_power=noise_level,
        is_energy=True,
        time_shift=False,
        channel_independent=False,
        plot=False)[0]
    time_vector = smooth_rir(energy_data, sampling_rate)[2]

    intersection_time_idx = np.rint(intersection_time * sampling_rate)

    energy_decay_curve = np.zeros([n_channels, n_samples])
    for idx_channel in range(0, n_channels):
        energy_decay_curve[
            idx_channel, :int(intersection_time_idx[idx_channel])] = \
                ra.schroeder_integration(
                    energy_data[
                        idx_channel, :int(intersection_time_idx[idx_channel])],
                    is_energy=True)

    if normalize:
        # Normalize the EDC...
        if not channel_independent:
            # ...by the first element of each channel.
            energy_decay_curve = (energy_decay_curve.T /
                                  energy_decay_curve[..., 0]).T
        else:
            # ...by the maximum first element of each channel.
            max_start_value = np.amax(energy_decay_curve[..., 0])
            energy_decay_curve /= max_start_value

    # Recover original data shape:
    energy_decay_curve = np.reshape(energy_decay_curve, data_shape)
    energy_decay_curve = np.squeeze(energy_decay_curve)

    if plot:
        plt.figure(figsize=(15, 3))
        plt.subplot(121)
        plt.plot(time_vector, 10*np.log10(energy_data.T))
        plt.xlabel('Time [s]')
        plt.ylabel('Squared IR [dB]')
        plt.subplot(122)
        plt.plot(time_vector[0:energy_decay_curve.shape[-1]], 10*np.log10(
            energy_decay_curve.T))
        plt.xlabel('Time [s]')
        plt.ylabel('EDC: Truncation method [dB]')
        plt.tight_layout()

    return energy_decay_curve


def energy_decay_curve_lundeby(
        data,
        sampling_rate,
        freq='broadband',
        noise_level='auto',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False):
    """ Lundeby et al. [#]_ proposed a correction term to prevent the truncation
    error. The missing signal energy from truncation time to infinity is
    estimated and added to the truncated integral.

    Parameters
    ----------
    data : ndarray, double
        The room impulse response with dimension [..., n_samples]
    sampling_rate: integer
        The sampling rate of the room impulse response.
    freq: integer OR string
        The frequency band. If set to 'broadband',
        the time window of the Lundeby-algorithm will not be set in dependence
        of frequency.
    noise_level: ndarray, double OR string
        If not specified, the noise level is calculated based on the last 10
        percent of the RIR. Otherwise specify manually for each channel
        as array.
    is_energy: boolean
        Defines, if the data is already squared.
    time_shift : boolean
        Defines, if the silence at beginning of the RIR should be removed.
    channel_independent : boolean
        Defines, if the time shift and normalizsation is done
        channel-independently or not.
    normalize : boolean
        Defines, if the energy decay curve should be normalized in the end
        or not.
    plot: Boolean
        Specifies, whether the results should be visualized or not.

    Returns
    -------
    energy_decay_curve: ndarray, double
        Returns the noise handeled edc.

    References
    ----------
    .. [#]  Lundeby, Virgran, Bietz and Vorlaender - Uncertainties of
            Measurements in Room Acoustics - ACUSTICA Vol. 81 (1995)
    """

    energy_data, n_channels, data_shape = preprocess_rir(
        data,
        is_energy=is_energy,
        time_shift=time_shift,
        channel_independent=channel_independent)
    n_samples = energy_data.shape[-1]
    intersection_time, late_reverberation_time, noise_estimation = \
        intersection_time_lundeby(
            energy_data,
            sampling_rate=sampling_rate,
            freq=freq,
            initial_noise_power=noise_level,
            is_energy=True,
            time_shift=False,
            channel_independent=False,
            plot=False)
    time_vector = smooth_rir(energy_data, sampling_rate)[2]

    energy_decay_curve = np.zeros([n_channels, n_samples])

    for idx_channel in range(0, n_channels):
        intersection_time_idx = np.argmin(
            np.abs(time_vector - intersection_time[idx_channel]))
        p_square_at_intersection = noise_estimation[idx_channel]

        # Calculate correction term according to DIN EN ISO 3382
        # TO-DO: check reference!
        correction = (p_square_at_intersection
                      * late_reverberation_time[idx_channel]
                      * (1 / (6*np.log(10)))
                      * sampling_rate)

        energy_decay_curve[idx_channel, :intersection_time_idx] = \
            ra.schroeder_integration(
                energy_data[idx_channel, :intersection_time_idx],
                is_energy=True)
        energy_decay_curve[idx_channel] += correction

    if normalize:
        # Normalize the EDC...
        if not channel_independent:
            # ...by the first element of each channel.
            energy_decay_curve = (
                energy_decay_curve.T /
                energy_decay_curve[..., 0]).T
        else:
            # ...by the maximum first element of each channel.
            max_start_value = np.amax(energy_decay_curve[..., 0])
            energy_decay_curve /= max_start_value

    energy_decay_curve[..., intersection_time_idx:] = np.nan

    if plot:
        plt.figure(figsize=(15, 3))
        plt.subplot(121)
        plt.plot(time_vector, 10*np.log10(energy_data.T))
        plt.xlabel('Time [s]')
        plt.ylabel('Squared IR [dB]')
        plt.subplot(122)
        plt.plot(time_vector[0:energy_decay_curve.shape[-1]], 10*np.log10(
            energy_decay_curve.T))
        plt.xlabel('Time [s]')
        plt.ylabel('Tr. EDC with correction [dB]')
        plt.tight_layout()

    # Recover original data shape:
    energy_decay_curve = np.reshape(energy_decay_curve, data_shape)
    energy_decay_curve = np.squeeze(energy_decay_curve)

    return energy_decay_curve


def energy_decay_curve_chu(
        data,
        sampling_rate,
        freq='broadband',
        noise_level='auto',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False):

    """ Implementation of the "subtraction of noise"-method after Chu [#]
    The noise level is estimated and subtracted from the impulse response
    before backward integration.

    Parameters
    ----------
    data : ndarray, double
        The room impulse response with dimension [..., n_samples]
    sampling_rate: integer
        The sampling rate of the room impulse response.
    freq: integer OR string
        The frequency band. If set to 'broadband',
        the time window of the Lundeby-algorithm will not be set in dependence
        of frequency.
    noise_level: ndarray, double OR string
        If not specified, the noise level is calculated based on the last 10
        percent of the RIR. Otherwise specify manually for each channel
        as array.
    is_energy: boolean
        Defines, if the data is already squared.
    time_shift : boolean
        Defines, if the silence at beginning of the RIR should be removed.
    channel_independent : boolean
        Defines, if the time shift and normalizsation is done
        channel-independently or not.
    normalize : boolean
        Defines, if the energy decay curve should be normalized in the end
        or not.
    plot: Boolean
        Specifies, whether the results should be visualized or not.

    Returns
    -------
    energy_decay_curve: ndarray, double
        Returns the noise handeled edc.

    References
    ----------
    .. [#]  W. T. Chu. “Comparison of reverberation measurements using
            Schroeder’s impulse method and decay-curve averaging method”.
            In: Journal of the Acoustical Society of America 63.5 (1978),
            pp. 1444–1450.
    """

    energy_data, n_channels, data_shape = preprocess_rir(
        data,
        is_energy=is_energy,
        time_shift=time_shift,
        channel_independent=channel_independent)

    subtracted = subtract_noise_from_squared_rir(
        energy_data,
        noise_level=noise_level)

    energy_decay_curve = ra.schroeder_integration(
        subtracted,
        is_energy=True)

    energy_decay_curve = np.atleast_2d(energy_decay_curve)

    if normalize:
        # Normalize the EDC...
        if not channel_independent:
            # ...by the first element of each channel.
            energy_decay_curve = (energy_decay_curve.T
                                  / energy_decay_curve[..., 0]).T
        else:
            # ...by the maximum first element of all channels.
            max_start_value = np.amax(energy_decay_curve[..., 0])
            energy_decay_curve /= max_start_value

    mask = energy_decay_curve <= 2*np.finfo(float).eps
    if np.any(mask):
        first_zero = np.nanargmax(mask, axis=-1)
        for channel in range(n_channels):
            energy_decay_curve[channel, first_zero[channel]:] = np.nan

    if plot:
        time_vector = (0.5+np.arange(0, energy_data.shape[-1]))/sampling_rate
        plt.figure(figsize=(15, 3))
        plt.subplot(131)
        plt.plot(time_vector, 10*np.log10(energy_data.T))
        plt.xlabel('Time [s]')
        plt.ylabel('Squared IR [dB]')
        plt.grid(True)
        plt.subplot(132)
        plt.plot(time_vector, 10*np.log10(subtracted.T))
        plt.xlabel('Time [s]')
        plt.ylabel('Noise subtracted IR [dB]')
        plt.grid(True)
        plt.subplot(133)
        plt.plot(time_vector, 10*np.log10(energy_decay_curve.T))
        plt.xlabel('Time [s]')
        plt.ylabel('Noise-handeled EDC [dB]')
        plt.grid(True)
        plt.tight_layout()

    # Recover original data shape:
    energy_decay_curve = np.reshape(energy_decay_curve, data_shape)
    energy_decay_curve = np.squeeze(energy_decay_curve)

    return energy_decay_curve


def energy_decay_curve_chu_lundeby(
        data,
        sampling_rate,
        freq='broadband',
        noise_level='auto',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False):
    """ This function combines Chu's and Lundeby's methods:
    The estimated noise level is subtracted before backward integration,
    the impulse response is truncated at the intersection time,
    and the correction for the truncation is applied [4]_, [5]_, [6]_

    Parameters
    ----------
    data : ndarray, double
        The room impulse response with dimension [..., n_samples]
    sampling_rate: integer
        The sampling rate of the room impulse response.
    freq: integer OR string
        The frequency band. If set to 'broadband',
        the time window of the Lundeby-algorithm will not be set in dependence
        of frequency.
    noise_level: ndarray, double OR string
        If not specified, the noise level is calculated based on the last 10
        percent of the RIR. Otherwise specify manually for each channel
        as array.
    is_energy: boolean
        Defines, if the data is already squared.
    time_shift : boolean
        Defines, if the silence at beginning of the RIR should be removed.
    channel_independent : boolean
        Defines, if the time shift and normalizsation is done
        channel-independently or not.
    normalize : boolean
        Defines, if the energy decay curve should be normalized in the end
        or not.
    plot: Boolean
        Specifies, whether the results should be visualized or not.

    Returns
    -------
    energy_decay_curve: ndarray, double
        Returns the noise handeled edc.

    References
    ----------
    .. [4]  Lundeby, Virgran, Bietz and Vorlaender - Uncertainties of
            Measurements in Room Acoustics - ACUSTICA Vol. 81 (1995)
    .. [5]  W. T. Chu. “Comparison of reverberation measurements using
            Schroeder’s impulse method and decay-curve averaging method”. In:
            Journal of the Acoustical Society of America 63.5 (1978),
            pp. 1444–1450.
    .. [6]  M. Guski, “Influences of external error sources on measurements of
            room acoustic parameters,” 2015.
    """

    energy_data, n_channels, data_shape = preprocess_rir(
        data,
        is_energy=is_energy,
        time_shift=time_shift,
        channel_independent=channel_independent)
    n_samples = energy_data.shape[-1]

    subtraction = subtract_noise_from_squared_rir(
        energy_data,
        noise_level=noise_level)

    intersection_time, late_reverberation_time, noise_level = \
        intersection_time_lundeby(
            energy_data,
            sampling_rate=sampling_rate,
            freq=freq,
            initial_noise_power=noise_level,
            is_energy=True,
            time_shift=False,
            channel_independent=False,
            plot=False)

    time_vector = smooth_rir(energy_data, sampling_rate)[2]
    energy_decay_curve = np.zeros([n_channels, n_samples])

    for idx_channel in range(0, n_channels):
        intersection_time_idx = np.argmin(np.abs(
            time_vector - intersection_time[idx_channel]))
        if noise_level == 'auto':
            p_square_at_intersection = estimate_noise_energy(
                energy_data[idx_channel], is_energy=True)
        else:
            p_square_at_intersection = noise_level[idx_channel]

        # calculate correction term according to DIN EN ISO 3382
        correction = (p_square_at_intersection
                      * late_reverberation_time[idx_channel]
                      * (1 / (6*np.log(10)))
                      * sampling_rate)

        energy_decay_curve[idx_channel, :intersection_time_idx] = \
            ra.schroeder_integration(
                subtraction[idx_channel, :intersection_time_idx],
                is_energy=True)
        energy_decay_curve[idx_channel] += correction

    if normalize:
        # Normalize the EDC...
        if not channel_independent:
            # ...by the first element of each channel.
            energy_decay_curve = (energy_decay_curve.T /
                                  energy_decay_curve[..., 0]).T
        else:
            # ...by the maximum first element of each channel.
            max_start_value = np.amax(energy_decay_curve[..., 0])
            energy_decay_curve /= max_start_value

    energy_decay_curve[..., intersection_time_idx:] = np.nan

    if plot:
        plt.figure(figsize=(15, 3))
        plt.subplot(131)
        plt.plot(time_vector, 10*np.log10(energy_data.T))
        plt.xlabel('Time [s]')
        plt.ylabel('Squared IR [dB]')
        plt.subplot(132)
        plt.plot(time_vector, 10*np.log10(subtraction.T))
        plt.xlabel('Time [s]')
        plt.ylabel('Noise subtracted IR [dB]')
        plt.subplot(133)
        plt.plot(time_vector[0:energy_decay_curve.shape[-1]], 10*np.log10(
            energy_decay_curve.T))
        plt.xlabel('Time [s]')
        plt.ylabel('Tr. EDC with corr. & subt. [dB]')
        plt.tight_layout()

    # Recover original data shape:
    energy_decay_curve = np.reshape(energy_decay_curve, data_shape)
    energy_decay_curve = np.squeeze(energy_decay_curve)

    return energy_decay_curve


def intersection_time_lundeby(
        data,
        sampling_rate,
        freq='broadband',
        initial_noise_power='auto',
        is_energy=False,
        time_shift=False,
        channel_independent=False,
        plot=False):

    """ This function uses the algorithm after Lundeby et al. [#]_ to calculate
    the intersection time, lundeby reverberation time, and noise level
    estimation.

    Parameters
    ----------
    data : ndarray, double
        The room impulse response with dimension [..., n_samples]
    sampling_rate: integer
        The sampling rate of the room impulse response.
    freq: integer OR string
        The frequency band. If set to 'broadband',
        the time window of the Lundeby-algorithm will not be set in dependence
        of frequency.
    noise_level: ndarray, double OR string
        If not specified, the noise level is calculated based on the last 10
        percent of the RIR. Otherwise specify manually for each channel
        as array.
    is_energy: boolean
        Defines, if the data is already squared.
    time_shift : boolean
        Defines, if the silence at beginning of the RIR should be removed.
    channel_independent : boolean
        Defines, if the time shift and normalizsation is done
        channel-independently or not.
    plot: Boolean
        Specifies, whether the results should be visualized or not.

    Returns
    -------
    intersection_time: ndarray, float
        Returns the Lundeby intersection time for each channel.
    reverberation_time: ndarray, float
        Returns the Lundeby reverberation time for each channel.
    noise_level: ndarray, float
        Returns the noise level estimation for each channel.

    References
    ----------
    .. [#]  Lundeby, Virgran, Bietz and Vorlaender - Uncertainties of
            Measurements in Room Acoustics - ACUSTICA Vol. 81 (1995)

    """
    # Define constants:
    # time intervals per 10 dB decay. Lundeby: 3...10
    n_intervals_per_10dB = 5
    # end of regression 5 ... 10 dB
    dB_above_noise = 10
    # Dynamic range 10 ... 20 dB
    use_dyn_range_for_regression = 20

    energy_data, n_channels, data_shape = preprocess_rir(
        data,
        is_energy=is_energy,
        time_shift=time_shift,
        channel_independent=channel_independent)

    if freq == "broadband":
        # broadband: use 30 ms windows sizes
        freq_dependent_window_time = 0.03
    else:
        freq_dependent_window_time = (800/freq+10) / 1000

    # (1) SMOOTH
    time_window_data, time_vector_window, time_vector = smooth_rir(
        energy_data, sampling_rate, freq_dependent_window_time)

    # (2) ESTIMATE NOISE
    if initial_noise_power == 'auto':
        noise_estimation = estimate_noise_energy(
            energy_data, is_energy=True)
    else:
        noise_estimation = initial_noise_power.copy()

    # (3) REGRESSION
    reverberation_time = np.zeros(n_channels)
    noise_level = np.zeros(n_channels)
    intersection_time = np.zeros(n_channels)
    noise_peak_level = np.zeros(n_channels)

    for idx_channel in range(0, n_channels):
        time_window_data_current_channel = time_window_data[idx_channel]
        start_idx = np.nanargmax(time_window_data_current_channel, axis=-1)
        try:
            stop_idx = (np.argwhere(10*np.log10(
                time_window_data_current_channel[start_idx+1:-1]) >
                    (10*np.log10(noise_estimation[idx_channel]) +
                        dB_above_noise))[-1, 0] + start_idx)
        except IndexError:
            raise Exception(
                'Regression failed: Low SNR. Estimation terminated.')

        dyn_range = np.diff(10*np.log10(np.take(
            time_window_data_current_channel, [start_idx, stop_idx])))

        if (stop_idx == start_idx) or dyn_range > -5:
            raise Exception(
                'Regression failed: Low SNR. Estimation terminated.')

        # regression_matrix*slope = edc
        regression_matrix = np.vstack((np.ones(
            [stop_idx-start_idx]), time_vector_window[start_idx:stop_idx]))
        slope = np.linalg.lstsq(
            regression_matrix.T,
            10*np.log10(time_window_data_current_channel[start_idx:stop_idx]),
            rcond=None)[0]

        if slope[1] == 0 or np.any(np.isnan(slope)):
            raise Exception(
                'Regression failed: T would be Inf, setting to 0. \
                    Estimation terminated.')

        regression_time = np.array(
            [time_vector_window[start_idx], time_vector_window[stop_idx]])
        regression_values = np.array(
            [10*np.log10(time_window_data[0, start_idx]),
             (10*np.log10(time_window_data[0, start_idx])
                + slope[1]*time_vector_window[stop_idx])])

        # (4) PRELIMINARY CROSSING POINT
        crossing_point = \
            (10*np.log10(noise_estimation[idx_channel]) - slope[0]) / slope[1]
        preliminary_crossing_point = crossing_point

        # (5) NEW LOCAL TIME INTERVAL LENGTH
        n_blocks_in_decay = (np.diff(
            10*np.log10(np.take(
                time_window_data_current_channel, [start_idx, stop_idx])))
            / -10 * n_intervals_per_10dB)

        n_samples_per_block = np.round(np.diff(np.take(
            time_vector_window,
            [start_idx, stop_idx])) / n_blocks_in_decay * sampling_rate)

        window_time = n_samples_per_block/sampling_rate

        # (6) AVERAGE
        time_window_data_current_channel, \
            time_vector_window_current_channel, \
            time_vector_current_channel = smooth_rir(
                energy_data[idx_channel], sampling_rate, window_time)
        time_window_data_current_channel = np.squeeze(
            time_window_data_current_channel)
        idx_max = np.nanargmax(time_window_data_current_channel)

        # high start value to enter while-loop
        old_crossing_point = 11+crossing_point
        loop_counter = 0

        while(True):
            # (7) ESTIMATE BACKGROUND LEVEL
            corresponding_decay = 10  # 5...10 dB
            idx_last_10_percent = np.round(
                time_window_data_current_channel.shape[-1]*0.9)
            idx_10dB_below_crosspoint = np.nanmax([1, np.round(
                ((crossing_point
                  - corresponding_decay / slope[1])
                 * sampling_rate / n_samples_per_block))])

            noise_estimation_current_channel = np.nanmean(
                time_window_data_current_channel[int(np.nanmin(
                    [idx_last_10_percent, idx_10dB_below_crosspoint])):])

            # (8) ESTIMATE LATE DECAY SLOPE
            try:
                start_idx_loop = np.argwhere(10*np.log10(
                    time_window_data_current_channel[idx_max:]) < (
                        10*np.log10(noise_estimation_current_channel)
                        + dB_above_noise
                        + use_dyn_range_for_regression
                        + idx_max))[0, 0]
            except TypeError:
                start_idx_loop = 0

            try:
                stop_idx_loop = (np.argwhere(10*np.log10(
                    time_window_data_current_channel[start_idx_loop+1:]) < (
                        10*np.log10(noise_estimation_current_channel)
                        + dB_above_noise))[0, 0]
                                 + start_idx_loop)
            except IndexError:
                raise Exception(
                    'Regression failed: Low SNR. Estimation terminated.')

            # regression_matrix*slope = edc
            regression_matrix = np.vstack((np.ones(
                [stop_idx_loop-start_idx_loop]),
                time_vector_window_current_channel[
                    start_idx_loop:stop_idx_loop]))

            slope = np.linalg.lstsq(
                regression_matrix.T,
                (10*np.log10(time_window_data_current_channel[
                    start_idx_loop:stop_idx_loop])),
                rcond=None)[0]

            if slope[1] >= 0:
                raise Exception(
                    'Regression did not work due, T would be Inf, \
                        setting to 0. Estimation was terminated.')

            # (9) FIND CROSSPOINT
            old_crossing_point = crossing_point
            crossing_point = ((10*np.log10(noise_estimation_current_channel)
                               - slope[0]) / slope[1])

            loop_counter = loop_counter + 1

            if (np.abs(old_crossing_point-crossing_point) < 0.01):
                break
            if loop_counter > 30:
                # TO-DO: Paper says 5 iterations are sufficient in all cases!
                warnings.warn(
                    "Lundeby algorithm was terminated after 30 iterations.")
                break

        reverberation_time[idx_channel] = -60/slope[1]
        noise_level[idx_channel] = noise_estimation_current_channel
        intersection_time[idx_channel] = crossing_point
        noise_peak_level[idx_channel] = 10 * np.log10(
            np.nanmax(time_window_data_current_channel[int(np.nanmin(
                [idx_last_10_percent, idx_10dB_below_crosspoint])):]))

    if plot:
        plt.figure(figsize=(15, 3))
        plt.subplot(131)
        max_data_db = np.nanmax(10*np.log10(energy_data))
        plt.plot(time_vector, 10*np.log10(energy_data.T))
        plt.xlabel('Time [s]')
        plt.ylabel('Squared RIR [dB]')
        plt.ylim(bottom=max_data_db-80+5, top=max_data_db+5)
        plt.grid(True)

        plt.subplot(132)
        plt.plot(time_vector_window, 10*np.log10(time_window_data.T))
        plt.xlabel('Time [s]')
        plt.ylabel('Smoothened RIR [dB]')
        plt.plot(
            time_vector_window[-int(np.round(time_window_data.shape[-1]/10))],
            10*np.log10(noise_estimation[0]),
            marker='o',
            label='noise estimation',
            linestyle='--',
            color='C1')
        plt.axhline(
                10*np.log10(noise_estimation[0]),
                color='C1',
                linestyle='--')
        plt.plot(
            regression_time,
            regression_values,
            marker='o',
            color='C2',
            label='regression')
        plt.plot(
            preliminary_crossing_point,
            10*np.log10(noise_estimation[idx_channel]),
            marker='o',
            color='C3',
            label='preliminary crosspoint')
        plt.legend()
        plt.grid(True)

        plt.subplot(133)
        # TO-DO: plot all channels?
        plt.plot(
            time_vector_window_current_channel,
            10*np.log10(time_window_data_current_channel))
        plt.xlabel('Time [s]')
        plt.ylabel('Final EDC Estimation [dB]')
        plt.plot(
            time_vector_window_current_channel[int(idx_10dB_below_crosspoint)],
            10*np.log10(noise_estimation[0]),
            marker='o',
            color='C1',
            label='Noise')
        plt.axhline(
            10*np.log10(noise_estimation[0]),
            color='C1',
            linestyle='--')
        plt.plot(
            crossing_point,
            10*np.log10(noise_estimation[0]),
            marker='o',
            color='C3',
            label='Final crosspoint',
            linestyle='--')
        plt.axvline(
            reverberation_time[0],
            label=('$T_{60} = '+str(np.round(reverberation_time[0], 3)) + '$'),
            color='k')
        plt.axvline(
            intersection_time[0],
            label=('Lundeby intersection time'),
            color='C3',
            linestyle='--')
        plt.tight_layout()
        plt.legend()
        plt.grid(True)

    return intersection_time, reverberation_time, noise_level



import numpy as np
import numpy.testing as npt
import pyrato.analytic as analytic


def test_analytic_shoebox_eigenfreqs():
    L = np.array([8, 5, 3])/10

    eigenfreqs, _ = \
        analytic.eigenfrequencies_rectangular_room_rigid(
            L,
            max_freq=1e3,
            speed_of_sound=343.9)

    f_n = np.array([
        0,
        5.731666666666666,
        3.439000000000000,
        6.684214522124330,
        6.877999999999999,
        8.953149545147662,
        2.149375000000000,
        6.121422683363956,
        4.055432639142833,
        7.021291666666667,
        7.206018102296510,
        9.207534939841542,
        4.298750000000000,
        7.164583333333334,
        5.505086063132890,
        7.947199213576930,
        8.110865278285665,
        9.931673491425187,
        6.448125000000000,
        8.627300782597231,
        7.307874999999999,
        9.287466812506130,
        9.427894781743429,
        8.597500000000000,
        9.259790885867778]) * 1e2

    npt.assert_allclose(eigenfreqs, f_n)


def test_analytic_shoebox_rir():
    L = np.array([8, 5, 3])/10
    source_pos = np.array([5, 3, 1.2])/10
    receiver_pos = [1, 0, 0]

    rir, _ = analytic.rectangular_room_rigid_walls(
        L,
        source_pos,
        receiver_pos,
        1,
        max_freq=1e3,
        n_samples=2**16,
        speed_of_sound=343.9)

    ref = np.loadtxt(
        './tests/data/analytic_rir_rigid_walls.csv',
        delimiter=',',
        skiprows=1)

    npt.assert_allclose(rir, ref)


def test_eigenfreq_impedance_1d_real():
    L = 8/10
    zeta = np.ones(2) * 1e10
    c = 343.9

    k = [0.1]
    k_max = 1e3*2*np.pi/c

    k_ns = analytic.eigenfrequencies_rectangular_room_1d(
        L, k, k_max, zeta
    )

    truth = np.array([
        1e-6,
        2.149375000000000,
        4.298750000000000,
        6.448125000000000,
        8.597500000000000]) * 1e2

    f_n = np.squeeze(c*k_ns.real/2/np.pi)
    npt.assert_allclose(f_n, truth, rtol=1e-3, atol=1e-3)


def test_eigenfreq_impedance_1d_real_jac():
    L = 8/10
    zeta = np.ones(2) * 1e10
    c = 343.9

    k = [0.1]
    k_max = 1e3*2*np.pi/c

    k_ns = analytic.eigenfrequencies_rectangular_room_1d(
        L, k, k_max, zeta, gradient=True
    )

    truth = np.array([
        1e-6,
        2.149375000000000,
        4.298750000000000,
        6.448125000000000,
        8.597500000000000]) * 1e2

    f_n = np.squeeze(c*k_ns.real/2/np.pi)
    npt.assert_allclose(f_n, truth, rtol=1e-3, atol=1e-3)


def test_analytic_shoebox_eigenfreqs_impedance_multi_k():
    L = np.array([8, 5, 3])/10
    zetas = np.ones((3, 2)) * 1e10

    c = 343.9

    k_max = 1e3*2*np.pi/c
    k = np.linspace(0, k_max*1.1, 2**10)

    k_ns, _ = analytic.eigenfrequencies_rectangular_room_impedance(
            L, k, k_max, zetas
        )

    truth = np.array([
        4.196434e-05,
        5.731666666666666,
        3.439000000000000,
        6.684214522124330,
        6.877999999999999,
        8.953149545147662,
        2.149375000000000,
        6.121422683363956,
        4.055432639142833,
        7.021291666666667,
        7.206018102296510,
        9.207534939841542,
        4.298750000000000,
        7.164583333333334,
        5.505086063132890,
        7.947199213576930,
        8.110865278285665,
        9.931673491425187,
        6.448125000000000,
        8.627300782597231,
        7.307874999999999,
        9.287466812506130,
        9.427894781743429,
        8.597500000000000,
        9.259790885867778]) * 1e2

    f_n = np.mean(np.squeeze(c*k_ns.real/2/np.pi), axis=-1)
    npt.assert_allclose(np.sort(f_n), np.sort(truth), atol=1e-3, rtol=1e-3)


def test_analytic_shoebox_eigenfreqs_impedance():
    L = np.array([8, 5, 3])/10
    zetas = np.ones((3, 2)) * 1e10

    c = 343.9

    k = [0.1]
    k_max = 1e3*2*np.pi/c

    k_ns, _ = analytic.eigenfrequencies_rectangular_room_impedance(
            L, k, k_max, zetas
        )

    truth = np.array([
        4.440943e-06,
        5.731666666666666,
        3.439000000000000,
        6.684214522124330,
        6.877999999999999,
        8.953149545147662,
        2.149375000000000,
        6.121422683363956,
        4.055432639142833,
        7.021291666666667,
        7.206018102296510,
        9.207534939841542,
        4.298750000000000,
        7.164583333333334,
        5.505086063132890,
        7.947199213576930,
        8.110865278285665,
        9.931673491425187,
        6.448125000000000,
        8.627300782597231,
        7.307874999999999,
        9.287466812506130,
        9.427894781743429,
        8.597500000000000,
        9.259790885867778]) * 1e2

    f_n = np.squeeze(c*k_ns.real/2/np.pi)
    npt.assert_allclose(np.sort(f_n), np.sort(truth), atol=1e-3, rtol=1e-3)


def test_analytic_eigenfrequencies_impedance_cplx():
    L = np.array([8, 5, 3])/10
    zetas = np.ones((3, 2)) * 1e10

    c = 343.9

    k_max = 1e3*2*np.pi/c
    k_min = 150*2*np.pi/c
    k = np.linspace(k_min, k_max*1.1, 2**10)

    k_ns, _ = analytic.eigenfrequencies_rectangular_room_impedance(
            L, k, k_max, zetas, only_normal=True
        )

    k_ns_x = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_x.csv',
        delimiter=',',
        dtype=complex)
    k_ns_y = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_y.csv',
        delimiter=',',
        dtype=complex)
    k_ns_z = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_z.csv',
        delimiter=',',
        dtype=complex)

    npt.assert_allclose(k_ns[0], k_ns_x, rtol=1e-6)
    npt.assert_allclose(k_ns[1], k_ns_y, rtol=1e-6)
    npt.assert_allclose(k_ns[2], k_ns_z, rtol=1e-6)


def test_analytic_eigenfrequencies_impedance_zeta15():
    L = np.array([8, 5, 3])/10
    zetas = np.ones((3, 2)) * 15

    c = 343.9

    k_max = 1e3*2*np.pi/c
    k_min = 150*2*np.pi/c
    k = np.linspace(k_min, k_max*1.1, 2**10)

    k_ns, _ = analytic.eigenfrequencies_rectangular_room_impedance(
            L, k, k_max, zetas, only_normal=True
        )

    k_ns_x = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_x_zeta15.csv',
        delimiter=',',
        dtype=complex)
    k_ns_y = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_y_zeta15.csv',
        delimiter=',',
        dtype=complex)
    k_ns_z = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_z_zeta15.csv',
        delimiter=',',
        dtype=complex)

    npt.assert_allclose(k_ns[0], k_ns_x, rtol=1e-6)
    npt.assert_allclose(k_ns[1], k_ns_y, rtol=1e-6)
    npt.assert_allclose(k_ns[2], k_ns_z, rtol=1e-6)


def test_analytic_pressure_shoebox_impedance():
    L = np.array([8, 5, 3])/10
    zetas = np.ones((3, 2)) * 1e10

    c = 343.9

    k_max = 1e3*2*np.pi/c
    k_min = 150*2*np.pi/c
    k = np.linspace(k_min, k_max*1.1, 2**10)

    k_ns_x = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_x.csv',
        delimiter=',',
        dtype=complex)
    k_ns_y = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_y.csv',
        delimiter=',',
        dtype=complex)
    k_ns_z = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_z.csv',
        delimiter=',',
        dtype=complex)

    k_ns = list((k_ns_x, k_ns_y, k_ns_z))

    mode_indices = np.loadtxt(
        'tests/data/analytic_impedance/mode_indices.csv',
        delimiter=',',
        dtype=int)

    r_R = np.array([3.3, 1.6, 1.8])/10
    r_S = np.array([5.3, 3.6, 1.2])/10

    p_x = analytic.pressure_modal_superposition(
        k, k*c, k_ns, mode_indices, r_R, r_S, L, zetas)

    truth = np.loadtxt(
        'tests/data/analytic_impedance/p_x.csv',
        delimiter=',',
        dtype=complex)

    npt.assert_allclose(p_x, truth, atol=1e-6, rtol=1e-6)


def test_analytic_pressure_shoebox_impedance_multi_R():
    L = np.array([8, 5, 3])/10
    zetas = np.ones((3, 2)) * 1e10

    c = 343.9

    k_max = 1e3*2*np.pi/c
    k_min = 150*2*np.pi/c
    k = np.linspace(k_min, k_max*1.1, 2**10)

    k_ns_x = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_x.csv',
        delimiter=',',
        dtype=complex)
    k_ns_y = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_y.csv',
        delimiter=',',
        dtype=complex)
    k_ns_z = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_z.csv',
        delimiter=',',
        dtype=complex)

    k_ns = list((k_ns_x, k_ns_y, k_ns_z))

    mode_indices = np.loadtxt(
        'tests/data/analytic_impedance/mode_indices.csv',
        delimiter=',',
        dtype=int)

    r_R = np.array([
        [3.3, 1.6, 1.8],
        [3.3, 1.6, 1.8]])/10
    r_S = np.array([5.3, 3.6, 1.2])/10

    p_x = analytic.pressure_modal_superposition(
        k, k*c, k_ns, mode_indices, r_R, r_S, L, zetas)

    truth = np.loadtxt(
        'tests/data/analytic_impedance/p_x.csv',
        delimiter=',',
        dtype=complex)

    truth = np.vstack((truth, truth))

    npt.assert_allclose(p_x, truth, atol=1e-6, rtol=1e-6)


def test_analytic_pressure_shoebox_impedance_zeta15():
    L = np.array([8, 5, 3])/10
    zetas = np.ones((3, 2)) * 15

    c = 343.9

    k_max = 1e3*2*np.pi/c
    k_min = 150*2*np.pi/c
    k = np.linspace(k_min, k_max*1.1, 2**10)

    k_ns_x = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_x_zeta15.csv',
        delimiter=',',
        dtype=complex)
    k_ns_y = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_y_zeta15.csv',
        delimiter=',',
        dtype=complex)
    k_ns_z = np.loadtxt(
        'tests/data/analytic_impedance/k_ns_z_zeta15.csv',
        delimiter=',',
        dtype=complex)

    k_ns = list((k_ns_x, k_ns_y, k_ns_z))

    mode_indices = np.loadtxt(
        'tests/data/analytic_impedance/mode_indices_zeta15.csv',
        delimiter=',',
        dtype=int)

    r_R = np.array([3.3, 1.6, 1.8])/10
    r_S = np.array([5.3, 3.6, 1.2])/10

    p_x = analytic.pressure_modal_superposition(
        k, k*c, k_ns, mode_indices, r_R, r_S, L, zetas)

    truth = np.loadtxt(
        'tests/data/analytic_impedance/p_x_zeta15.csv',
        delimiter=',',
        dtype=complex)

    npt.assert_allclose(p_x, truth, atol=1e-4, rtol=1e-4)


def test_analytic_shoebox_spec_impedance():
    L = np.array([8, 5, 3])/10
    zetas = np.ones((3, 2)) * 100

    c = 343.9

    r_R = np.array([3.3, 1.6, 1.8])/10
    r_S = np.array([5.3, 3.6, 1.2])/10

    f_max = 1e3

    samplingrate = 2200
    n_samples = 2**10

    rir, spec, k_ns = analytic.rectangular_room_impedance(
        L,
        r_S,
        r_R,
        zetas,
        f_max,
        samplingrate,
        c,
        n_samples,
        remove_cavity_mode=False)

    truth_rtf = np.load('tests/data/analytic_rtf_impedance.npy')
    truth_rir = np.load('tests/data/analytic_rir_impedance.npy')

    npt.assert_allclose(spec, truth_rtf, atol=1e-2, rtol=1e-2)
    npt.assert_allclose(rir, truth_rir, atol=1e-2, rtol=1e-2)


def test_analytic_shoebox_spec_impedance_no_cavity_mode():
    L = np.array([8, 5, 3])/10
    zetas = np.ones((3, 2)) * 100

    c = 343.9

    r_R = np.array([3.3, 1.6, 1.8])/10
    r_S = np.array([5.3, 3.6, 1.2])/10

    f_max = 1e3

    samplingrate = 2200
    n_samples = 2**10

    rir, spec, k_ns = analytic.rectangular_room_impedance(
        L,
        r_S,
        r_R,
        zetas,
        f_max,
        samplingrate,
        c,
        n_samples,
        remove_cavity_mode=True)

    truth_rtf = np.load('tests/data/analytic_rtf_impedance_no_cav.npy')
    truth_rir = np.load('tests/data/analytic_rir_impedance_no_cav.npy')

    npt.assert_allclose(spec, truth_rtf, atol=1e-2, rtol=1e-2)
    npt.assert_allclose(rir, truth_rir, atol=1e-2, rtol=1e-2)



# -*- coding: utf-8 -*-

import os

import numpy as np
import numpy.testing as npt
import pytest
from numpy import genfromtxt

import pyrato.dsp as dsp

test_data_path = os.path.join(os.path.dirname(__file__), 'test_data')


def test_start_ir_insufficient_snr():
    n_samples = 2**9
    ir = np.zeros(n_samples, dtype=float)
    ir[20] = 1

    snr = 15

    noise = np.random.randn(n_samples)
    noise = noise / np.sqrt(np.mean(np.abs(noise**2))) * 10**(-snr/20)
    ir_noise = ir + noise

    with pytest.raises(ValueError):
        dsp.find_impulse_response_start(ir_noise)


def test_start_ir():
    n_samples = 2**10
    ir = np.zeros(n_samples)

    snr = 60

    noise = np.random.randn(n_samples) * 10**(-snr/20)

    start_sample = 24
    ir[start_sample] = 1

    start_sample_est = dsp.find_impulse_response_start(ir)
    assert start_sample_est == start_sample - 1

    ir_awgn = ir + noise
    start_sample_est = dsp.find_impulse_response_start(ir_awgn)
    assert start_sample_est == start_sample - 1


def test_start_ir_thresh():
    n_samples = 2**10
    ir = np.zeros(n_samples)

    start_sample = 24
    ir[start_sample] = 1
    ir[start_sample-4:start_sample] = 10**(-5/10)

    start_sample_est = dsp.find_impulse_response_start(ir, threshold=20)
    assert start_sample_est == start_sample - 4 - 1


def test_start_ir_multidim():
    n_samples = 2**10
    n_channels = 3
    ir = np.zeros((n_channels, n_samples))

    snr = 60

    noise = np.random.randn(n_channels, n_samples) * 10**(-snr/20)

    start_sample = [24, 5, 43]
    ir[[0, 1, 2], start_sample] = 1

    ir_awgn = ir + noise
    start_sample_est = dsp.find_impulse_response_start(ir_awgn)

    npt.assert_allclose(start_sample_est, np.array(start_sample) - 1)

    ir = np.zeros((2, n_channels, n_samples))
    noise = np.random.randn(2, n_channels, n_samples) * 10**(-snr/20)

    start_sample_1 = [24, 5, 43]
    ir[0, [0, 1, 2], start_sample_1] = 1
    start_sample_2 = [14, 12, 16]
    ir[1, [0, 1, 2], start_sample_2] = 1

    start_samples = np.vstack((start_sample_1, start_sample_2))

    ir_awgn = ir + noise
    start_sample_est = dsp.find_impulse_response_start(ir_awgn)

    npt.assert_allclose(start_sample_est, start_samples - 1)


def test_time_shift_right():
    shift_samples = 10
    n_samples = 2**9
    ir = np.zeros(n_samples, dtype=float)
    ir[20] = 1

    ir_truth = np.zeros(n_samples, dtype=float)
    ir_truth[20+shift_samples] = 1
    ir_shifted = dsp.time_shift(ir, shift_samples)

    npt.assert_allclose(ir_shifted, ir_truth)


def test_time_shift_left():
    shift_samples = 10
    n_samples = 2**9
    ir = np.zeros(n_samples, dtype=float)
    ir[20] = 1

    ir_truth = np.zeros(n_samples, dtype=float)
    ir_truth[20-shift_samples] = 1
    ir_shifted = dsp.time_shift(ir, -shift_samples)

    npt.assert_allclose(ir_shifted, ir_truth)


def test_time_shift_non_circular_right():
    shift_samples = 10
    n_samples = 2**9
    ir = np.zeros(n_samples, dtype=float)
    ir[20] = 1

    ir_truth = np.zeros(n_samples, dtype=float)
    ir_truth[20+shift_samples] = 1
    ir_truth[:shift_samples] = np.nan
    ir_shifted = dsp.time_shift(ir, shift_samples, circular_shift=False)

    npt.assert_allclose(ir_shifted, ir_truth, equal_nan=True)


def test_time_shift_non_circular_left():
    shift_samples = 10
    n_samples = 2**9
    ir = np.zeros(n_samples, dtype=float)
    ir[20] = 1

    ir_truth = np.zeros(n_samples, dtype=float)
    ir_truth[20-shift_samples] = 1
    ir_truth[n_samples-shift_samples:] = np.nan
    ir_shifted = dsp.time_shift(ir, -shift_samples, circular_shift=False)

    npt.assert_allclose(ir_shifted, ir_truth, equal_nan=True)


def test_time_shift_multitim():
    shift_samples = 10
    n_samples = 2**10
    n_channels = 3
    ir = np.zeros((n_channels, n_samples))

    start_sample = [24, 5, 43]
    ir[[0, 1, 2], start_sample] = 1

    ir_truth = np.zeros((n_channels, n_samples), dtype=float)
    start_sample_truth = [24+shift_samples, 5+shift_samples, 43+shift_samples]
    ir_truth[[0, 1, 2], start_sample_truth] = 1

    ir_shifted = dsp.time_shift(ir, shift_samples)

    npt.assert_allclose(ir_shifted, ir_truth)

    ir_truth = np.zeros((n_channels, n_samples), dtype=float)
    start_sample_truth = [24-shift_samples, 5-shift_samples, 43-shift_samples]
    ir_truth[[0, 1, 2], start_sample_truth] = 1

    ir_shifted = dsp.time_shift(ir, -shift_samples)

    npt.assert_allclose(ir_shifted, ir_truth)


def test_time_shift_multitim_multishift():
    shift_samples = [10, 2, 4]
    n_samples = 40
    n_channels = 3
    ir = np.zeros((n_channels, n_samples), dtype=float)

    start_sample = [24, 5, 13]
    ir[[0, 1, 2], start_sample] = 1

    ir_truth = np.zeros((n_channels, n_samples), dtype=float)
    start_sample_truth = [
        24+shift_samples[0],
        5+shift_samples[1],
        13+shift_samples[2]]
    ir_truth[[0, 1, 2], start_sample_truth] = 1

    ir_shifted = dsp.time_shift(ir, shift_samples)

    npt.assert_allclose(ir_shifted, ir_truth)

    ir_truth = np.zeros((n_channels, n_samples), dtype=float)
    start_sample_truth = [
        24-shift_samples[0],
        5-shift_samples[1],
        13-shift_samples[2]]
    ir_truth[[0, 1, 2], start_sample_truth] = 1

    ir_shifted = dsp.time_shift(ir, -np.array(shift_samples, dtype=int))

    npt.assert_allclose(ir_shifted, ir_truth)


def test_start_room_impulse_response():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')

    actual = dsp.find_impulse_response_start(rir, threshold=20)

    expected = 0

    npt.assert_allclose(actual, expected)


def test_start_room_impulse_response_shfted(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')

    rir_shifted = np.roll(rir, 128, axis=-1)
    actual = dsp.find_impulse_response_start(rir_shifted, threshold=20)

    expected = 128

    npt.assert_allclose(actual, expected)


def test_start_ir_thresh_invalid():
    n_samples = 2**10
    ir = np.zeros(n_samples)

    start_sample = 24
    ir[start_sample] = 1
    # ir[start_sample-4:start_sample] = 10**(-10/10)
    ir[0:start_sample] = 10**(-5/10)

    start_sample_est = dsp.find_impulse_response_start(ir, threshold=20)
    assert start_sample_est == 0


def test_start_ir_thresh_invalid_osci():
    n_samples = 2**10
    ir = np.zeros(n_samples)

    start_sample = 24
    ir[start_sample] = 1
    ir[start_sample-4:start_sample] = 10**(-30/10)
    ir[0:start_sample-4] = 10**(-5/10)

    start_sample_est = dsp.find_impulse_response_start(ir, threshold=20)
    assert start_sample_est == 0


def test_max_ir():
    n_samples = 2**10
    ir = np.zeros(n_samples)

    snr = 60

    noise = np.random.randn(n_samples) * 10**(-snr/20)

    start_sample = 24
    ir[start_sample] = 1

    start_sample_est = dsp.find_impulse_response_maximum(ir)
    assert start_sample_est == start_sample

    ir_awgn = ir + noise
    start_sample_est = dsp.find_impulse_response_maximum(ir_awgn)
    assert start_sample_est == start_sample



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Tests for reverberation time related things. """
from pytest import raises

import numpy as np
import numpy.testing as npt

import pyrato as ra


def test_rt_from_edc():
    times = np.linspace(0, 1.5, 2**9)
    m = -60
    edc = times * m
    edc_exp = 10**(edc/10)
    RT = 1.
    TX = ['T20', 'T30', 'T40', 'T50', 'T60', 'LDT', 'EDT']
    for T in TX:
        RT_est = ra.reverberation_time_energy_decay_curve(edc_exp, times, T=T)
        npt.assert_allclose(RT_est, RT)


def test_rt_from_edc_error():
    times = np.linspace(0, 1.5, 2**9)
    m = -60
    edc = times * m
    edc_exp = 10**(edc/10)
    T = 'Bla'

    with raises(ValueError, match='is not a valid interval.'):
        ra.reverberation_time_energy_decay_curve(edc_exp, times, T=T)
    # npt.assert_allclose(RT_est, RT)



# -*- coding: utf-8 -*-
from itertools import count

import numpy as np
from scipy import optimize


def eigenfrequencies_rectangular_room_rigid(
        dimensions, max_freq, speed_of_sound):
    """Calculate the eigenfrequencies of a rectangular room with rigid walls.

    Parameters
    ----------
    dimensions : double, ndarray
        The dimensions of the room in the form [L_x, L_y, L_z]
    max_freq : double
        The maximum frequency to consider for the calculation of the
        eigenfrequencies of the room
    speed_of_sound : double, optional (343.9)
        The speed of sound

    Returns
    -------
    f_n : double, ndarray
        The eigenfrequencies of the room
    n : int, ndarray
        The modal index

    References
    ----------
    ..  [2] H. Kuttruff, Room acoustics, pp. 64-66, 4th Ed. Taylor & Francis,
        2009.
    """
    c = speed_of_sound
    L = np.asarray(dimensions)
    L_x = dimensions[0]
    L_y = dimensions[1]
    L_z = dimensions[2]
    f_max = max_freq

    n_modes = 0
    n_x_max = int(np.floor(2*f_max/c * L_x)) + 1
    for n_x in range(0, n_x_max):
        n_y_max = int(np.floor(np.real(
            np.sqrt((2*f_max/c)**2 - (n_x/L_x)**2) * L_y))) + 1
        for n_y in range(0, n_y_max):
            n_modes += int(np.floor(np.real(
                np.sqrt(
                    (2*f_max/c)**2 - (n_x/L_x)**2 - (n_y/L_y)**2
                ) * L_z))) + 1

    n = np.zeros((3, n_modes))

    idx = 0
    n_x_max = int(np.floor(2*f_max/c * L_x)) + 1
    for n_x in range(0, n_x_max):
        n_y_max = int(np.floor(np.real(
            np.sqrt((2*f_max/c)**2 - (n_x/L_x)**2) * L_y))) + 1
        for n_y in range(0, n_y_max):
            n_z_max = int(np.floor(np.real(
                np.sqrt(
                    (2*f_max/c)**2 - (n_x/L_x)**2 - (n_y/L_y)**2
                ) * L_z))) + 1

            idx_end = idx + n_z_max
            n[0, idx:idx_end] = n_x
            n[1, idx:idx_end] = n_y
            n[2, idx:idx_end] = np.arange(0, n_z_max)

            idx += n_z_max

    f_n = c/2*np.sqrt(np.sum((n/L[np.newaxis].T)**2, axis=0))

    return f_n, n


def rectangular_room_rigid_walls(dimensions,
                                 source,
                                 receiver,
                                 reverberation_time,
                                 max_freq,
                                 samplingrate=44100,
                                 speed_of_sound=343.9,
                                 n_samples=2**18):
    r"""Calculate the transfer function of a rectangular room based on the
    analytic model as given in [2]_ . The model is based on the solution
    for a room with rigid walls. The damping of the modes is included as
    a damping in the medium, not as a damping caused by the boundary.
    Consequently, all modes share the same damping factor calculated from
    the reverberation time as :math:`\delta = \frac{3\log(10)}{T_{60}}`.

    Parameters
    ----------
    dimensions : double, ndarray
        The dimensions of the room in the form [L_x, L_y, L_z]
    source : double, array
        The source position in Cartesian coordinates [x, y, z]
    receiver : double, ndarray
        The receiver position in Cartesian coordinates [x, y, z]
    max_freq : double
        The maximum frequency to consider for the calculation of the
        eigenfrequencies of the room
    samplingrate : int
        The sampling rate
    speed_of_sound : double, optional (343.9)
        The speed of sound
    n_samples : int
        number of samples for the calculation

    Returns
    -------
    rir : ndarray, double
        The room impulse response
    eigenfrequencies: ndarray, double
        The eigenfrequencies for which the room impulse response was
        calculated

    References
    ----------
    ..  [2] H. Kuttruff, Room acoustics, pp. 64-66, 4th Ed. Taylor & Francis,
        2009.


    """
    delta_n_raw = 3*np.log(10)/reverberation_time

    c = speed_of_sound
    L = np.asarray(dimensions)
    L_x = dimensions[0]
    L_y = dimensions[1]
    L_z = dimensions[2]
    source = np.asarray(source)
    receiver = np.asarray(receiver)

    f_n, n = eigenfrequencies_rectangular_room_rigid(
        dimensions, max_freq, speed_of_sound)

    coeff_receiver = \
        np.cos(np.pi*n[0]*receiver[0]/L_x) * \
        np.cos(np.pi*n[1]*receiver[1]/L_y) * \
        np.cos(np.pi*n[2]*receiver[2]/L_z)
    coeff_source = \
        np.cos(np.pi*n[0]*source[0]/L_x) * \
        np.cos(np.pi*n[1]*source[1]/L_y) * \
        np.cos(np.pi*n[2]*source[2]/L_z)

    K_n = np.prod(L) * 0.5**(np.sum(n > 0, axis=0))
    factor = c**2 / K_n
    coeff = coeff_source * coeff_receiver * factor

    coeff[0] = 0.

    freqs = np.fft.rfftfreq(n_samples, d=1 / samplingrate)
    n_bins = freqs.size
    omega = 2*np.pi*freqs
    omega_n = 2*np.pi*f_n
    omega_squared = omega**2

    transfer_function = np.zeros(n_bins, complex)
    for om_n, coeff_n in zip(omega_n, coeff):
        den = omega_squared - delta_n_raw**2 - om_n**2 - 2*1j*delta_n_raw*omega
        transfer_function += (coeff_n/den)

    rir = np.fft.irfft(transfer_function, n=n_samples)
    return rir, f_n


def transcendental_equation_eigenfrequencies_impedance(k_n, k, L, zeta):
    """The transcendental equation to be solved for the estimation of the
    complex eigenfrequencies of the rectangular room with uniform impedances.
    This function is intended as the cost function for solving for the roots.

    Parameters
    ----------
    k_n : array, double
        The real and imaginary part of the complex eigenfrequency
    k : double
        The real valued wave number
    L : double
        The room dimension
    zeta : array, double
        The normalized specific impedance

    Returns
    -------
    func : array, double
        The real and imaginary part of the transcendental equation
    """
    k_n_real = k_n[0]
    k_n_imag = k_n[1]

    k_n_complex = k_n_real + 1j*k_n_imag

    left = np.tan(k_n_complex*L)
    right = \
        (1j*k*L*np.sum(zeta)) / (k_n_complex*L * (np.prod(zeta) +
                                 (k*L)**2/(k_n_complex*L)**2))
    func = left - right

    return [func.real, func.imag]


def transcendental_equation_eigenfrequencies_impedance_newton(k_n, k, L, zeta):
    r"""The transcendental equation to be solved for the estimation of the
    complex eigenfrequencies of the rectangular room with uniform impedances.
    This function is intended as the cost function for solving for the roots.

    .. math::

        \frac{ i k \left(\zeta_{0} + \zeta_{L}\right)}{k_{n}
        \left(\frac{k^{2}}{k_{n}^{2}} + \zeta_{0} \zeta_{L}\right)} +
        \tan{\left(L k_{n} \right)} = 0

    Parameters
    ----------
    k_n : complex
        The complex eigenvalue
    k : double
        The real valued wave number
    L : double
        The room dimension
    zeta : array, double
        The normalized specific impedance

    Returns
    -------
    func : array, double
        The complex transcendental equation
    """
    left = np.tan(k_n*L)
    zeta_prod = zeta[0]*zeta[1]
    zeta_sum = zeta[0]+zeta[1]

    right = \
        (1j*k*zeta_sum) / (k_n * (zeta_prod + (k/k_n)**2))
    func = left - right

    return func


def gradient_trancendental_equation_eigenfrequencies_impedance(
        k_n, k, L, zeta):
    r"""The gradient of the  transcendental equation for the estimation of the
    complex eigenfrequencies of the rectangular room with uniform impedances.
    This function is intended as the analytic jacobian function for solving
    for the roots of the transcendental equation.

    .. math::

        L \left(\tan^{2}{\left(L k_{n} \right)} + 1\right) -
        \frac{2 i k^{3} \left(\zeta_{0} + \zeta_{L}\right)}{k_{n}^{4}
        \left(\frac{k^{2}}{k_{n}^{2}} + \zeta_{0} \zeta_{L}\right)^{2}} +
        \frac{ i k \left(\zeta_{0} + \zeta_{L}\right)}{k_{n}^{2}
        \left(\frac{k^{2}}{k_{n}^{2}} + \zeta_{0} \zeta_{L}\right)} = 0

    Parameters
    ----------
    k_n : complex
        The complex eigenvalue
    k : double
        The real valued wave number
    L : double
        The room dimension
    zeta : array, double
        The normalized specific impedance

    Returns
    -------
    func : array, double
        The complex transcendental equation
    """
    zeta_prod = zeta[0]*zeta[1]
    zeta_sum = zeta[0]+zeta[1]

    tan = L * (np.tan(L * k_n)**2 + 1)
    denom = (k_n**2 * ((k/k_n)**2 + zeta_prod))
    left = k**3 / denom**2
    right = k / denom

    d_k_n = tan + (-2*left + right)*zeta_sum*1j
    return d_k_n


def initial_solution_transcendental_equation(k, L, zeta):
    """ Initial solution to the transcendental equation for the complex
    eigenfrequencies of the rectangular room with uniform impedance at
    the boundaries. This will approximate the zeroth order mode.

    Parameters
    ----------
    k : array, double
        Wave number array

    Returns
    -------
    k_0 : array, complex
        The complex zero order eigenfrequency
    """
    zeta_0 = zeta[0]
    zeta_L = zeta[1]
    k_0 = 1/L*np.sqrt(-(k*L)**2/zeta_0/zeta_L + 1j*k*L*(1/zeta_0+1/zeta_L))

    return k_0


def eigenfrequencies_rectangular_room_1d(
        L_l, ks, k_max, zeta, gradient=True):
    """Estimates the complex eigenvalues in the wavenumber domain for one
    dimension by numerically solving for the roots of the transcendental
    equation. A initial approximation to the zeroth order mode is applied to
    improve the conditioning of the problem.

    Parameters
    ----------
    L_l : double
        The dimension in m
    ks : array, double
        The wave numbers for which the eigenvalues are to be solved.
    k_max : double
        The real part of the largest eigenvalue. This solves as a stopping
        criterion independent from the real wave number k.
    zeta : array, double
        The normalized specific impedance on the boundaries.
    gradient : boolean, optional (True)
        Use the analytic gradient of the transcendental equation instead
        of an numerical approximation in the solver

    Returns
    -------
    k_ns : array, complex
        The complex eigenvalues for each wavenumber

    Note
    ----
    This function assumes that the real part of the largest eigenvalue may be
    calculated using the approximation for rigid walls.

    """
    ks = np.atleast_1d(ks)
    n_l_max = int(np.ceil(k_max/np.pi*L_l))

    if gradient:
        fprime = gradient_trancendental_equation_eigenfrequencies_impedance
    else:
        fprime = False

    k_ns_l = np.zeros((n_l_max, len(ks)), dtype=complex)
    k_n_init = initial_solution_transcendental_equation(ks[0], L_l, zeta)
    for idx_k, k in enumerate(ks):
        idx_n = 0
        while k_n_init.real < k_max:
            args_costfun = (k, L_l, zeta)
            kk_n = optimize.newton(
                transcendental_equation_eigenfrequencies_impedance_newton,
                k_n_init,
                fprime=fprime,
                args=args_costfun)
            if kk_n.real > k_max:
                break
            else:
                k_ns_l[idx_n, idx_k] = kk_n
                k_n_init = (kk_n*L_l + np.pi) / L_l
                idx_n += 1

        k_n_init = k_ns_l[0, idx_k]

    return k_ns_l


def normal_eigenfrequencies_rectangular_room_impedance(
        L, ks, k_max, zeta):
    r"""Caller function for the eigenvalue estimation of all room dimensions.
    See the function `eigenfrequencies_rectangular_room_impedance` or
    `eigenfrequencies_rectangular_room_1d` for more information.

    Parameters
    ----------
    L : array, double
        The dimensions in m
    ks : array, double
        The wave numbers for which the eigenvalues are to be solved.
    k_max : double
        The real part of the largest eigenvalue. This solves as a stopping
        criterion independent from the real wave number k.
    zeta : array, double
        The normalized specific impedance on the boundaries.

    Returns
    -------
    k_ns : list, complex
        List of arrays with the complex eigenvalues for each wavenumber and
        each room dimension.
    """
    k_ns = []
    for dim, L_l, zeta_l in zip(count(), L, zeta):
        k_ns_l = eigenfrequencies_rectangular_room_1d(
            L_l, ks, k_max, zeta_l)
        k_ns.append(k_ns_l)
    return k_ns


def eigenfrequencies_rectangular_room_impedance(
        L, ks, k_max, zeta, only_normal=False):
    r"""Estimates the complex eigenvalues in the wavenumber domain for a
    rectangular room with arbitrary uniform impedances on the boundary by
    numerically solving for the roots of the transcendental equation.
    A initial approximation to the zeroth order mode is applied to
    improve the conditioning of the problem. The eigenvalues corresponding to
    tangential and oblique modes are calculated from the eigenvalues of the
    respective axial modes. Resulting eigenvalues with a real part larger than
    k_max will be discarded.

    Parameters
    ----------
    L : array, double
        The dimensions in m
    ks : array, double
        The wave numbers for which the eigenvalues are to be solved.
    k_max : double
        The real part of the largest eigenvalue. This solves as a stopping
        criterion independent from the real wave number k.
    zeta : array, double
        The normalized specific impedance on the boundaries.
    only_normal : boolean, optional (False)
        Only return the eigenvalues corresponding to the axial modes.
        The mode indices will still contain the indices for all modes in
        the defined frequency range. The complete set of eigenvalues
        can be calculated as
        :math:`k_n = \sqrt{ k_{n,x}^2+k_{n,y}^2+k_{n,z}^2 }`.

    Returns
    -------
    k_ns : array, complex
        The complex eigenvalues for each wavenumber
    mode_indices : array, integer
        The wave number indices of respective eigenvalues.

    Note
    ----
    Eigenvalues smaller for a wave number :math:`k < 0.02` will be replaced by
    the value for the closest larger wave number to ensure finding the root.

    """
    ks = np.atleast_1d(ks)
    mask = ks >= 0.02
    ks_search = ks[mask]
    k_ns = normal_eigenfrequencies_rectangular_room_impedance(
        L, ks_search, k_max, zeta
    )
    for idx in range(0, len(L)):
        k_ns[idx] = np.hstack((
            np.tile(k_ns[idx][:, 0], (np.sum(~mask), 1)).T,
            k_ns[idx]))

    n_z = np.arange(0, k_ns[2].shape[0])
    n_y = np.arange(0, k_ns[1].shape[0])
    n_x = np.arange(0, k_ns[0].shape[0])

    combs = np.meshgrid(n_x, n_y, n_z)
    perms = np.array(combs).T.reshape(-1, 3)

    kk_ns = np.sqrt(
        k_ns[0][perms[:, 0]]**2 +
        k_ns[1][perms[:, 1]]**2 +
        k_ns[2][perms[:, 2]]**2)

    mask_perms = (kk_ns[:, -1].real < k_max)

    mask_bc = np.broadcast_to(
        np.atleast_2d(mask_perms).T,
        (len(mask_perms), len(ks)))

    if only_normal:
        kk_ns = k_ns
    else:
        kk_ns = kk_ns[mask_bc].reshape(-1, len(ks))

    mode_indices = perms[mask_bc[:, 0]]

    return kk_ns, mode_indices


def mode_function_impedance(position, eigenvalue, phase):
    r""" The modal function for a room with boundary impedances [4]_ .

    .. math::

        p_{n,i}(x_i) = \cosh(x_i k_{n,i} + \phi_{n,i})

    Parameters
    ----------
    position : ndarray, double, (3,)
        The position in Cartesian coordinates
    eigenvalue : ndarray, complex, (3, N, n_bins)
        The N complex eigenvalues in x,y,z coordinates for each
        frequency bin (wavenumber)
    phase : ndarray, complex, (3, N, n_bins)
        The phase shift introduced by the boundary impedance

    Returns
    -------
    p_n : ndarray, complex, (3, N, n_bins)
        The modal function

    References
    ----------
    ..  [4] M. Nolan and J. L. Davy, “Two definitions of the inner product of
        modes and their use in calculating non-diffuse reverberant sound
        fields,” The Journal of the Acoustical Society of America, vol.
        145, no. 6, pp. 3330–3340, Jun. 2019.
    """
    return np.cosh(1j*eigenvalue * position + phase)


def pressure_modal_superposition(
        ks, omegas, k_ns, mode_indices, r_R, r_S, L, zeta):
    r""" Calculate modal composition for a rectangular room with arbitrary
    boundary impedances.

    Parameters
    ----------
    ks : ndarray, double
        The wave number array
    omegas : ndarray, double
        The angular frequency array :math:`omega`
    k_ns : list, complex
        List containing the complex eigenvalues for each dimension and
        wavenumber
    r_R : ndarray, double, (3, n_receivers)
        The receiver positions in Cartesian coordinates
    r_S : ndarray, double, (3)
        The source position in Cartesian coordinates
    L : ndarray, double, (3,)
        The room dimensions in meters
    zeta : ndarray, double, (3, 2)
        The normalized impedance :math:`\zeta_i = \frac{Z_i}{\rho_o c}`
        for each wall.

    Returns
    -------
    spec : ndarray, complex
        The complex sum of all mode functions corresponding to the eigenvalues
        `k_ns`

    """

    zeta_0 = zeta[:, 0]
    r_R = np.atleast_2d(r_R)

    kk_ns = np.sqrt(
        k_ns[0][mode_indices[:, 0]]**2 +
        k_ns[1][mode_indices[:, 1]]**2 +
        k_ns[2][mode_indices[:, 2]]**2)

    k_ns_xyz = np.array([
        k_ns[0][mode_indices[:, 0]],
        k_ns[1][mode_indices[:, 1]],
        k_ns[2][mode_indices[:, 2]]
        ])

    phi = np.arctanh(ks/(zeta_0 * k_ns_xyz.T).T)
    K_n_sc = \
        np.sinh(1j * k_ns_xyz.T * L) * \
        np.cosh(1j * k_ns_xyz.T * L + 2*phi.T)

    K_n = np.prod((L/2 * (1 + 1/(1j*k_ns_xyz.T * L) * K_n_sc)).T, axis=0)
    denom = K_n * (kk_ns**2 - ks**2)

    p_ns_s = np.prod(mode_function_impedance(r_S, k_ns_xyz.T, phi.T).T, axis=0)

    spec = np.zeros((r_R.shape[0], ks.size), dtype=complex)
    for idx_R in range(r_R.shape[0]):
        p_ns_r = np.prod(
            mode_function_impedance(
                r_R[idx_R, :], k_ns_xyz.T, phi.T).T,
            axis=0)

        nom = 1j*omegas*1.2*p_ns_r*p_ns_s

        spec[idx_R, :] = np.sum(nom / denom, axis=0)

    spec = np.squeeze(spec)

    return spec


def rectangular_room_impedance(
        L,
        r_S,
        r_R,
        normalized_impedance,
        max_freq,
        samplingrate=44100,
        c=343.9,
        n_samples=2**12,
        remove_cavity_mode=False):
    r""" Calculate the room impulse response and room transfer function for a
    rectangular room with arbitrary boundary impedances.

    Parameters
    ----------
    L : ndarray, double, (3,)
        The room dimensions in meters
    r_S : ndarray, double, (3)
        The source position in Cartesian coordinates
    r_R : ndarray, double, (3, n_receivers)
        The receiver positions in Cartesian coordinates
    normalized_impedance : ndarray, double, (3, 2)
        The normalized impedance :math:`\zeta_i = \frac{Z_i}{\rho_o c}`
        for each wall.
    max_freq : double
        The highest frequency to be considered for the estimation of the
        eigenfrequencies.
    samplingrate : int, 44100
        The samplingrate
    c : float, 343.9
        The speed of sound in m/s
    n_samples : int, 2**12
        The number of samples for which the RIR is calculated
    remove_cavity_mode : boolean, False
        When true, the cavity mode (0, 0, 0) will be removed before summation
        of all modes

    Returns
    -------
    rir : ndarray, double, (n_receivers, n_samples)
        The room impulse response
    rtf : ndarray, double, (n_receivers, n_bins)
        The room transfer function in the frequency domain
    eigenvalues : ndarray, complex
        The complex eigenvalues in the form
        :math:`k_n = \omega_n / c + i \delta_n`
    """

    zeta = normalized_impedance
    freqs = np.fft.rfftfreq(n_samples, 1/samplingrate)
    ks = 2*np.pi*freqs/c

    k_max = max_freq*2*np.pi/c
    k_ns, mode_indices = eigenfrequencies_rectangular_room_impedance(
        L, ks, k_max, zeta, only_normal=True)

    if remove_cavity_mode:
        mask = np.prod(np.array([0, 0, 0]) == mode_indices, axis=-1) == 1
        mode_indices = mode_indices[~mask]

    spectrum = pressure_modal_superposition(
        ks, freqs*2*np.pi, k_ns, mode_indices, r_R, r_S, L, zeta)

    rir = np.fft.irfft(spectrum)

    k_ns_xyz = np.array([
        k_ns[0][mode_indices[:, 0]],
        k_ns[1][mode_indices[:, 1]],
        k_ns[2][mode_indices[:, 2]]
        ])

    return rir, spectrum, k_ns_xyz



# -*- coding: utf-8 -*-

"""Main module."""

import re
import numpy as np
import matplotlib.pyplot as plt


def reverberation_time_energy_decay_curve(
        energy_decay_curve,
        times,
        T='T20',
        normalize=True,
        plot=False):
    """Estimate the reverberation time from a given energy decay curve according
    to the ISO standard 3382 [#]_.

    Parameters
    ----------
    energy_decay_curve : ndarray, double
        Energy decay curve. The time needs to be the arrays last dimension.
    times : ndarray, double
        Time vector corresponding to each sample of the EDC.
    T : 'T20', 'T30', 'T40', 'T50', 'T60', 'EDT', 'LDT'
        Decay interval to be used for the reverberation time extrapolation. EDT
        corresponds to the early decay time extrapolated from the interval
        [0, -10] dB, LDT corresponds to the late decay time extrapolated from
        the interval [-25, -35] dB.
    normalize : bool, True
        Normalize the EDC to the steady state energy level
    plot : bool, False
        Plot the estimated extrapolation line for visual inspection of the
        results.

    Returns
    -------
    reverberation_time : double
        The reverberation time

    References
    ----------
    .. [#]  ISO 3382, Acoustics - Measurement of the reverberation time of
            rooms with reference to other acoustical parameters.

    """
    intervals = [20, 30, 40, 50, 60]

    if T == 'EDT':
        upper = -0.1
        lower = -10.1
    elif T == 'LDT':
        upper = -25.
        lower = -35.
    else:
        try:
            (int(re.findall(r'\d+', T)[0]) in intervals)
        except IndexError:
            raise ValueError(
                "{} is not a valid interval for the regression.".format(T))

        upper = -5
        lower = -np.double(re.findall(r'\d+', T)) + upper

    if normalize:
        energy_decay_curve /= energy_decay_curve[0]

    edc_db = 10*np.log10(np.abs(energy_decay_curve))

    idx_upper = np.nanargmin(np.abs(upper - edc_db))
    idx_lower = np.nanargmin(np.abs(lower - edc_db))

    A = np.vstack(
        [times[idx_upper:idx_lower], np.ones(idx_lower - idx_upper)]).T
    gradient, const = np.linalg.lstsq(
        A, edc_db[..., idx_upper:idx_lower], rcond=None)[0]

    reverberation_time = -60 / gradient

    if plot:
        plt.figure()
        plt.plot(
            times,
            edc_db,
            label='edc')
        plt.plot(
            times,
            times * gradient + const,
            label='regression',
            linestyle='-.')
        ax = plt.gca()
        ax.set_ylim((-95, 5))

        reverberation_time = -60 / gradient

        ax.set_xlim((-0.05, 2*reverberation_time))
        plt.grid(True)
        plt.legend()
        ax.set_ylabel('EDC [dB]')
        ax.set_xlabel('Time [s]')

    return reverberation_time


def schroeder_integration(impulse_response, is_energy=False):
    r"""Calculate the Schroeder integral of a room impulse response [#]_. The
    result is the energy decay curve for the given room impulse response.

    .. math:

        \langle e^2(t) \rangle = N\cdot \int_{t}^{\infty} h^2(\tau)
        \mathrm{d} \tau

    Parameters
    ----------
    impulse_response : ndarray, double
        Room impulse response as array
    is_energy : boolean, optional
        Whether the input represents energy data or sound pressure values.

    Returns
    -------
    energy_decay_curve : ndarray, double
        The energy decay curve

    References
    ----------
    .. [#]  M. R. Schroeder, “New Method of Measuring Reverberation Time,”
            The Journal of the Acoustical Society of America, vol. 37, no. 6,
            pp. 1187–1187, 1965.

    """
    if not is_energy:
        data = np.abs(impulse_response)**2
    else:
        data = impulse_response.copy()

    ndim = data.ndim
    data = np.atleast_2d(data)
    energy_decay_curve = np.fliplr(np.nancumsum(np.fliplr(data), axis=-1))

    if ndim < energy_decay_curve.ndim:
        energy_decay_curve = np.squeeze(energy_decay_curve)

    return energy_decay_curve


def energy_decay_curve_analytic(
        surfaces, alphas, volume, times, source=None,
        receiver=None, method='eyring', c=343.4, frequency=None,
        air_absorption=True):
    """Calculate the energy decay curve analytically by using Eyring's or
    Sabine's equation [#]_.

    Parameters
    ----------
    surfaces : ndarray, double
        Surface areas of all surfaces in the room
    alphas : ndarray, double
        Absorption coefficients corresponding to each surface
    volume : double
        Room volume
    times : ndarray, double
        Time vector for which the decay curve is calculated
    source : Coordinates
        Coordinate object with the source coordinates
    receiver : Coordinates
        Coordinate object with the receiver coordinates
    method : 'eyring', 'sabine'
        Use either Eyring's or Sabine's equation
    c : double
        Speed of sound
    frequency : double, optional
        Center frequency of the respective octave band. This is only used for
        the air absorption calculation.

    Returns
    -------
    energy_decay_curve : ndarray, double
        The energy decay curve

    References
    ----------
    .. [#]  H. Kuttruff, Room acoustics, 4th Ed. Taylor & Francis, 2009.

    """

    alphas = np.asarray(alphas)
    surfaces = np.asarray(surfaces)
    surface_room = np.sum(surfaces)
    alpha_mean = np.sum(surfaces*alphas) / surface_room

    if air_absorption:
        m = air_attenuation_coefficient(frequency)
    else:
        m = 0

    if all([source, receiver]):
        dist_source_receiver = np.linalg.norm(
            source.cartesian - receiver.cartesian)
        delay_direct = dist_source_receiver / c
    else:
        delay_direct = 0

    if method == 'eyring':
        energy_decay_curve = np.exp(
            -c*(times - delay_direct) *
            ((-surface_room * np.log(1 - alpha_mean) / 4 / volume) + m))
    elif method == 'sabine':
        energy_decay_curve = np.exp(
            -c*(times - delay_direct) *
            ((surface_room * alpha_mean / 4 / volume) + m))
    else:
        raise ValueError("The method has to be either 'eyring' or 'sabine'.")

    return energy_decay_curve


def air_attenuation_coefficient(
        frequency,
        temperature=20,
        humidity=50,
        atmospheric_pressure=101325):
    """Calculate the attenuation coefficient m for the absorption caused
     by friction with the surrounding air.

    Parameters
    ----------
    frequency : double
        The frequency for which the attenuation coefficient is calculated.
        When processing in fractional octave bands use the center frequency.
    temperature : double
        Temperature in degrees Celsius.
    humidity : double
        Humidity in percent.
    atmospheric_pressure : double
        Atmospheric pressure.

    Returns
    -------
    attenuation_coefficient : double
        The resulting attenuation coefficient.

    """
    # room temperature in Kelvin
    t_K = temperature + 273.16
    p_ref_kPa = 101.325
    p_kPa = atmospheric_pressure/1000.0

    # determine molar concentration of water vapor
    tmp = (
        (10.79586 * (1.0 - (273.16/t_K))) -
        (5.02808 * np.log10((t_K/273.16))) +
        (1.50474 * 0.0001 * (1.0 - 10.0 ** (-8.29692*((t_K/273.16) - 1.0)))) +
        (0.42873 * 0.001 * (-1.0 + 10.0 ** (-4.76955*(1.0 - (273.16/t_K))))) -
        2.2195983)

    # molar concentration water vapor in percent
    molar_water_vapor = (humidity * 10.0 ** tmp) / (p_kPa/p_ref_kPa)

    # determine relaxation frequencies of oxygen and nitrogen
    relax_oxygen = ((p_kPa/p_ref_kPa) * (24.0 + (
            4.04 * 10000.0 * molar_water_vapor * (
                (0.02 + molar_water_vapor) / (0.391 + molar_water_vapor)))))

    relax_nitrogen = ((p_kPa/p_ref_kPa) * (
        (t_K / 293.16) ** (-0.5)) *
        (9.0 + 280.0 * molar_water_vapor * np.exp(
            -4.17 * (((t_K / 293.16) ** (-0.3333333)) - 1.0))))

    # Neper/m -> dB/m
    air_abs_coeff = ((frequency**2 * (
        (1.84 * 10.0**(-11.0) * (p_ref_kPa / p_kPa) * (t_K/293.16)**0.5) +
        ((t_K/293.16)**(-2.5) * (
            (1.278 * 0.01 * np.exp(-2239.1/t_K) / (
                relax_oxygen + (frequency**2)/relax_oxygen)) +
            (1.068 * 0.1 * np.exp((-3352.0/t_K) / (
                relax_nitrogen + (frequency**2)/relax_nitrogen))))))
            ) * 20.0 / np.log(10.0) / (np.log10(np.exp(1.0)) * 10.0))

    return air_abs_coeff



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    The generate_test_data module provides the functionallity to generate
    adequate test data for automatic testing.
"""

__author__ = "Johannes Imort"
__version__ = "0.1.0"
__maintainer__ = "Johannes Imort"
__email__ = "johannes.imort@rwth-aachen.de"
__status__ = "Development"

# %%
import numpy as np
import pyrato
from pyrato import analytic


sampling_rate = 3000
room_dimensions = [1, 1, 1]
src_pos = [0.2, 0.3, 0.4]
rec_pos = [0.5, 0.6, 0.7]
t_60 = 1
max_freq = 1000
n_samples = 2**13

rir_1 = analytic.rectangular_room_rigid_walls(
    dimensions=room_dimensions, source=src_pos, receiver=rec_pos, reverberation_time=t_60, max_freq=max_freq,
    samplingrate=sampling_rate, speed_of_sound=343.9, n_samples=n_samples)[0]

rir_2 = analytic.rectangular_room_rigid_walls(
    dimensions=room_dimensions, source=src_pos, receiver=rec_pos, reverberation_time=t_60*2, max_freq=max_freq,
    samplingrate=sampling_rate, speed_of_sound=343.9, n_samples=n_samples*2)[0]

rir_array = np.zeros(([2, rir_2.size]))

psnr = 50

rir_1 /= np.amax(np.abs(rir_1))
rir_2 /= np.amax(np.abs(rir_2))

rir_array[0,:rir_1.shape[0]] = rir_1
rir_array[1,:] = rir_2

# fix the seed
np.random.seed(1)
noise = np.random.normal(0, 1, rir_2.size)

rir_array += 10**(-(psnr-10) / 20) * noise
# %%

noise_energy_1D = pyrato.estimate_noise_energy(
    rir_array[0], interval=[0.9, 1.0], is_energy=False)
noise_energy_2D = pyrato.estimate_noise_energy(
    rir_array, interval=[0.9, 1.0], is_energy=False)

preprocessing_1D = pyrato.edc.preprocess_rir(
    rir_array[0], is_energy=False, time_shift=False, channel_independent=False)
preprocessing_2D = pyrato.edc.preprocess_rir(
    rir_array, is_energy=False, time_shift=False, channel_independent=False)

preprocessing_time_shift_1D = pyrato.edc.preprocess_rir(
    rir_array[0], is_energy=False, time_shift=True, channel_independent=False)
preprocessing_time_shift_2D = pyrato.edc.preprocess_rir(
    rir_array, is_energy=False, time_shift=True, channel_independent=False)

preprocessing_time_shift_channel_independent_1D = pyrato.edc.preprocess_rir(
    rir_array[0], is_energy=False, time_shift=True, channel_independent=True)
preprocessing_time_shift_channel_independent_2D = pyrato.edc.preprocess_rir(
    rir_array, is_energy=False, time_shift=True, channel_independent=True)

smoothed_rir_1D = pyrato.edc.smooth_rir(rir_array[0], sampling_rate, smooth_block_length=0.075)
smoothed_rir_2D = pyrato.edc.smooth_rir(rir_array, sampling_rate, smooth_block_length=0.075)

substracted_1D = pyrato.edc.subtract_noise_from_squared_rir(rir_array[0]**2)
substracted_2D = pyrato.edc.subtract_noise_from_squared_rir(rir_array**2)

edc_truncation_1D = pyrato.energy_decay_curve_truncation(
    rir_array[0], sampling_rate, freq='broadband', is_energy=False, time_shift=True,
    channel_independent=False, normalize=True)
edc_truncation_2D = pyrato.energy_decay_curve_truncation(
    rir_array, sampling_rate, freq='broadband', is_energy=False, time_shift=True,
    channel_independent=False, normalize=True)

edc_lundeby_1D = pyrato.energy_decay_curve_lundeby(
    rir_array[0], sampling_rate, freq='broadband', is_energy=False, time_shift=True,
    channel_independent=False, normalize=True, plot=False)
edc_lundeby_2D = pyrato.energy_decay_curve_lundeby(
    rir_array, sampling_rate, freq='broadband', is_energy=False, time_shift=True,
    channel_independent=False, normalize=True, plot=False)

edc_lundeby_chu_1D = pyrato.energy_decay_curve_chu_lundeby(
    rir_array[0], sampling_rate, freq='broadband', is_energy=False, time_shift=True,
    channel_independent=False, normalize=True, plot=False)
edc_lundeby_chu_2D = pyrato.energy_decay_curve_chu_lundeby(
    rir_array, sampling_rate, freq='broadband', is_energy=False, time_shift=True,
    channel_independent=False, normalize=True, plot=False)

edc_chu_1D = pyrato.energy_decay_curve_chu(
    rir_array[0], sampling_rate, freq='broadband', is_energy=False, time_shift=True,
    channel_independent=False, normalize=True, plot=False)
edc_chu_2D = pyrato.energy_decay_curve_chu(
    rir_array, sampling_rate, freq='broadband', is_energy=False, time_shift=True,
    channel_independent=False, normalize=True, plot=False)

intersection_time_1D = pyrato.intersection_time_lundeby(
    rir_array[0], sampling_rate, freq='broadband', is_energy=False, time_shift=False,
    channel_independent=False, plot=False)
intersection_time_2D = pyrato.intersection_time_lundeby(
    rir_array, sampling_rate, freq='broadband', is_energy=False, time_shift=False,
    channel_independent=False, plot=False)

# noise_energy_from_edc_1D = pyrato.edc.estimate_noise_energy_from_edc(
#     edc_lundeby_chu_1D, intersection_time_1D[0], sampling_rate)
# noise_energy_from_edc_2D = pyrato.edc.estimate_noise_energy_from_edc(
#     edc_lundeby_chu_2D, intersection_time_2D[0], sampling_rate)

# %%


np.savetxt("analytic_rir_psnr50_1D.csv", rir_array[0], delimiter=",")
np.savetxt("analytic_rir_psnr50_2D.csv", rir_array, delimiter=",")

np.savetxt("noise_energy_1D.csv", noise_energy_1D, delimiter=",")
np.savetxt("noise_energy_2D.csv", noise_energy_2D, delimiter=",")

np.savetxt("preprocessing_1D.csv", preprocessing_1D[0], delimiter=",")
np.savetxt("preprocessing_2D.csv", preprocessing_2D[0], delimiter=",")

np.savetxt("preprocessing_time_shift_1D.csv", preprocessing_time_shift_1D[0], delimiter=",")
np.savetxt("preprocessing_time_shift_2D.csv", preprocessing_time_shift_2D[0], delimiter=",")

np.savetxt("preprocessing_time_shift_channel_independent_1D.csv", preprocessing_time_shift_channel_independent_1D[0], delimiter=",")
np.savetxt("preprocessing_time_shift_channel_independent_2D.csv", preprocessing_time_shift_channel_independent_2D[0], delimiter=",")

np.savetxt("smoothed_rir_1D.csv", smoothed_rir_1D[0], delimiter=",")
np.savetxt("smoothed_rir_2D.csv", smoothed_rir_2D[0], delimiter=",")

np.savetxt("substracted_1D.csv", substracted_1D, delimiter=",")
np.savetxt("substracted_2D.csv", substracted_2D, delimiter=",")

np.savetxt("edc_truncation_1D.csv", edc_truncation_1D, delimiter=",")
np.savetxt("edc_truncation_2D.csv", edc_truncation_2D, delimiter=",")

np.savetxt("edc_lundeby_1D.csv", edc_lundeby_1D, delimiter=",")
np.savetxt("edc_lundeby_2D.csv", edc_lundeby_2D, delimiter=",")

np.savetxt("edc_lundeby_chu_1D.csv", edc_lundeby_chu_1D, delimiter=",")
np.savetxt("edc_lundeby_chu_2D.csv", edc_lundeby_chu_2D, delimiter=",")

np.savetxt("edc_chu_1D.csv", edc_chu_1D, delimiter=",")
np.savetxt("edc_chu_2D.csv", edc_chu_2D, delimiter=",")

np.savetxt("intersection_time_1D.csv", intersection_time_1D, delimiter=",")
np.savetxt("intersection_time_2D.csv", intersection_time_2D, delimiter=",")

#     np.savetxt("noise_energy_from_edc_1D.csv", noise_energy_from_edc_1D, delimiter=",")
#     np.savetxt("noise_energy_from_edc_2D.csv", noise_energy_from_edc_2D, delimiter=",")



# -*- coding: utf-8 -*-

"""Signal processing related functions."""

import warnings

import numpy as np
import scipy.signal as spsignal


def find_impulse_response_start(
        impulse_response,
        threshold=20,
        noise_energy='auto'):
    """Find the first sample of an impulse response in a accordance with the
    ISO standard ISO 3382 [#]_. The start sample is identified as the first
    sample that varies significantly from the noise floor but still has a level
    of at least 20 dB below the maximum of the impulse response. The function
    further tries to consider oscillations before the time below the threshold
    value.

    Parameters
    ----------
    impulse_response : ndarray, double
        The impulse response
    threshold : double, optional
        Threshold according to ISO3382 in dB

    Returns
    -------
    start_sample : int
        Sample at which the impulse response starts

    Note
    ----
    The function tries to estimate the SNR in the IR based on the signal energy
    in the last 10 percent of the IR.

    References
    ----------
    .. [#]  ISO 3382-1:2009-10, Acoustics - Measurement of the reverberation
            time of rooms with reference to other acoustical parameters. pp. 22
    """
    ir_squared = np.abs(impulse_response)**2

    mask_start = int(0.9*ir_squared.shape[-1])
    if noise_energy == 'auto':
        mask = np.arange(mask_start, ir_squared.shape[-1])
        noise = np.mean(np.take(ir_squared, mask, axis=-1), axis=-1)
    else:
        noise = noise_energy

    max_sample = np.argmax(ir_squared, axis=-1)
    max_value = np.max(ir_squared, axis=-1)

    if np.any(max_value < 10**(threshold/10) * noise) or \
            np.any(max_sample > mask_start):
        raise ValueError("The SNR is lower than the defined threshold. Check \
                if this is a valid impulse response with sufficient SNR.")

    start_sample_shape = max_sample.shape
    n_samples = ir_squared.shape[-1]
    ir_squared = np.reshape(ir_squared, (-1, n_samples))
    n_channels = ir_squared.shape[0]
    max_sample = np.reshape(max_sample, n_channels)
    max_value = np.reshape(max_value, n_channels)

    start_sample = max_sample.copy()
    for idx in range(n_channels):
        # Only look for the start sample if the maximum index is bigger than 0
        if start_sample[idx] > 0:
            ir_before_max = ir_squared[idx, :max_sample[idx]+1] \
                / max_value[idx]
            # Last value before peak lower than the peak/threshold
            idx_last_below_thresh = np.argwhere(
                ir_before_max < 10**(-threshold/10))
            if idx_last_below_thresh.size > 0:
                start_sample[idx] = idx_last_below_thresh[-1]
            else:
                start_sample[idx] = 0
                warnings.warn(
                    'No values below threshold found before the maximum value,\
                    defaulting to 0')

            idx_6dB_above_threshold = np.argwhere(
                ir_before_max[:start_sample[idx]+1] >
                10**((-threshold+6)/10))
            if idx_6dB_above_threshold.size > 0:
                idx_6dB_above_threshold = int(idx_6dB_above_threshold[0])
                tmp = np.argwhere(
                    ir_before_max[:idx_6dB_above_threshold+1] <
                    10**(-threshold/10))
                if tmp.size == 0:
                    start_sample[idx] = 0
                    warnings.warn(
                        'Oscillations detected in the impulse response. \
                        No clear starting sample found, defaulting to 0')
                else:
                    start_sample[idx] = tmp[-1]

    start_sample = np.reshape(start_sample, start_sample_shape)

    return np.squeeze(start_sample)


def find_impulse_response_maximum(
        impulse_response,
        threshold=20,
        noise_energy='auto'):
    """Find the maximum of an impulse response as argmax(h(t)).
    Performs an initial SNR check according to a defined threshold level in dB.

    Parameters
    ----------
    impulse_response : ndarray, double
        The impulse response
    threshold : double, optional
        Threshold SNR value in dB

    Returns
    -------
    max_sample : int
        Sample at which the impulse response starts

    Note
    ----
    The function tries to estimate the SNR in the IR based on the signal energy
    in the last 10 percent of the IR.

    """
    ir_squared = np.abs(impulse_response)**2

    mask_start = int(0.9*ir_squared.shape[-1])
    if noise_energy == 'auto':
        mask = np.arange(mask_start, ir_squared.shape[-1])
        noise = np.mean(np.take(ir_squared, mask, axis=-1), axis=-1)
    else:
        noise = noise_energy

    max_sample = np.argmax(ir_squared, axis=-1)
    max_value = np.max(ir_squared, axis=-1)

    if np.any(max_value < 10**(threshold/10) * noise) or \
            np.any(max_sample > mask_start):
        raise ValueError("The SNR is lower than the defined threshold. Check \
                if this is a valid impulse response with sufficient SNR.")

    return np.squeeze(max_sample)


def time_shift(signal, n_samples_shift, circular_shift=True, keepdims=False):
    """Shift a signal in the time domain by n samples.
    This function will perform a circular shift by default, inherently
    assuming that the signal is periodic. Use the option `circular_shift=False`
    to pad with nan values instead.

    Notes
    -----
    This function is primarily intended to be used when processing impulse
    responses.

    Parameters
    ----------
    signal : ndarray, float
        Signal to be shifted
    n_samples_shift : integer
        Number of samples by which the signal should be shifted. A negative
        number of samples will result in a left-shift, while a positive
        number of samples will result in a right shift of the signal.
    circular_shift : bool, True
        Perform a circular or non-circular shift. If a non-circular shift is
        performed, the data will be padded with nan values at the respective
        beginning or ending of the data, corresponding to the number of samples
        the data is shifted.
    keepdims : bool, False
        Do not squeeze the data before returning.

    Returns
    -------
    shifted_signal : ndarray, float
        Shifted input signal

    """
    n_samples_shift = np.asarray(n_samples_shift, dtype=int)
    if np.any(signal.shape[-1] < n_samples_shift):
        msg = "Shifting by more samples than length of the signal."
        if circular_shift:
            warnings.warn(msg, UserWarning)
        else:
            raise ValueError(msg)

    signal = np.atleast_2d(signal)
    n_samples = signal.shape[-1]
    signal_shape = signal.shape
    signal = np.reshape(signal, (-1, n_samples))
    n_channels = np.prod(signal.shape[:-1])

    if n_samples_shift.size == 1:
        n_samples_shift = np.broadcast_to(n_samples_shift, n_channels)
    elif n_samples_shift.size == n_channels:
        n_samples_shift = np.reshape(n_samples_shift, n_channels)
    else:
        raise ValueError("The number of shift samples has to match the number \
            of signal channels.")

    shifted_signal = signal.copy()
    for channel in range(n_channels):
        shifted_signal[channel, :] = \
            np.roll(
                shifted_signal[channel, :],
                n_samples_shift[channel],
                axis=-1)
        if not circular_shift:
            if n_samples_shift[channel] < 0:
                # index is negative, so index will reference from the
                # end of the array
                shifted_signal[channel, n_samples_shift[channel]:] = np.nan
            else:
                # index is positive, so index will reference from the
                # start of the array
                shifted_signal[channel, :n_samples_shift[channel]] = np.nan

    shifted_signal = np.reshape(shifted_signal, signal_shape)
    if not keepdims:
        shifted_signal = np.squeeze(shifted_signal)

    return shifted_signal


def center_frequencies_octaves():
    """Return the octave center frequencies according to the IEC 61260:1:2014
    standard.

    Returns
    -------
    frequencies : ndarray, float
        Octave center frequencies

    """
    nominal = np.array([31.5, 63, 125, 250, 500, 1e3,
                        2e3, 4e3, 8e3, 16e3], dtype=np.float)
    indices = _frequency_indices(nominal, 1)
    exact = exact_center_frequencies_fractional_octaves(indices, 1)

    return nominal, exact


def center_frequencies_third_octaves():
    """Return the third octave center frequencies according
    to the ICE 61260:1:2014 standard.

    Returns
    -------
    frequencies : ndarray, float
        third octave center frequencies

    """
    nominal = np.array([
        25, 31.5, 40, 50, 63, 80, 100, 125, 160,
        200, 250, 315, 400, 500, 630, 800, 1000,
        1250, 1600, 2000, 2500, 3150, 4000, 5000,
        6300, 8000, 10000, 12500, 16000, 20000], dtype=np.float)

    indices = _frequency_indices(nominal, 3)
    exact = exact_center_frequencies_fractional_octaves(indices, 3)

    return nominal, exact


def exact_center_frequencies_fractional_octaves(indices, num_fractions):
    """Returns the exact center frequencies for fractional octave bands
    according to the IEC 61260:1:2014 standard.

    octave ratio
    .. G = 10^{3/10}

    center frequencies
    .. f_m = f_r G^{x/b}
    .. f_m = f_e G^{(2x+1)/(2b)}

    where b is the number of octave fractions, f_r is the reference frequency
    chosen as 1000Hz and x is the index of the frequency band.

    Parameters
    ----------
    indices : array
        The indices for which the center frequencies are calculated.
    num_fractions : 1, 3
        The number of octave fractions. 1 returns octave center frequencies,
        3 returns third octave center frequencies.

    Returns
    -------
    frequencies : ndarray, float
        center frequencies of the fractional octave bands

    """

    reference_freq = 1e3
    octave_ratio = 10**(3/10)

    iseven = np.mod(num_fractions, 2) == 0
    if ~iseven:
        exponent = (indices/num_fractions)
    else:
        exponent = ((2*indices + 1) / num_fractions / 2)

    return reference_freq * octave_ratio**exponent


def _frequency_indices(frequencies, num_fractions):
    """Return the indices for fractional octave filters.

    Parameters
    ----------
    frequencies : array
        The nominal frequencies for which the indices for exact center
        frequency calculation are to be calculated.
    num_fractions : 1, 3
        Number of fractional bands

    Returns
    -------
    indices : array
        The indices for exact center frequency calculation.

    """
    reference_freq = 1e3
    octave_ratio = 10**(3/10)

    iseven = np.mod(num_fractions, 2) == 0
    if ~iseven:
        indices = np.around(
            num_fractions * np.log(frequencies/reference_freq)
            / np.log(octave_ratio))
    else:
        indices = np.around(
            2.0*num_fractions *
            np.log(frequencies/reference_freq) / np.log(octave_ratio) - 1)/2

    return indices


def filter_fractional_octave_bands(
        signal, samplingrate, num_fractions,
        freq_range=(20.0, 20e3), order=6):
    """Apply a fractional octave filter to a signal.
    Filter bank implementation using second order sections of butterworth
    filters for increased numeric accuracy and stability.

    Parameters
    ----------
    signal : ndarray
        input signal to be filtered
    samplingrate : integer
        samplingrate of the signal
    num_fractions : integer
        number of octave fractions
    order : integer, optional
        order of the butterworth filter

    Returns
    -------
    signal_filtered : ndarray
        Signal filtered into fractional octave bands. The array has a new axis
        with dimension corresponding to the number of frequency bands:
        [num_fractions, *signal.shape]

    """

    if num_fractions not in (1, 3):
        raise ValueError("This currently supports only octave and third \
                octave band filters.")

    octave_ratio = 10**(3/10)

    if num_fractions == 1:
        nominal, exact = center_frequencies_octaves()
    else:
        nominal, exact = center_frequencies_third_octaves()

    mask_frequencies = (nominal > freq_range[0]) & (nominal < freq_range[1])

    nominal = nominal[mask_frequencies]
    exact = exact[mask_frequencies]

    signal_out_shape = (exact.size,) + signal.shape
    signal_out = np.broadcast_to(signal, signal_out_shape).copy()

    for band in range(exact.size):
        freq_upper = exact[band] * octave_ratio**(1/2/num_fractions)
        freq_lower = exact[band] * octave_ratio**(-1/2/num_fractions)

        # normalize interval such that the Nyquist frequency is 1
        Wn = np.array([freq_lower, freq_upper]) / samplingrate * 2
        # in case the upper frequency limit is above Nyquist, use a highpass
        if Wn[-1] > 1:
            warnings.warn('Your upper frequency limit [{}] is above the \
                Nyquist frequency. Using a highpass filter instead of a \
                bandpass'.format(freq_upper))
            Wn = Wn[0]
            btype = 'highpass'
        else:
            btype = 'bandpass'
        sos = spsignal.butter(order, Wn, btype=btype, output='sos')
        signal_out[band, :] = spsignal.sosfilt(sos, signal)

    return signal_out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    The test_edc_noise_handling module provides the functionality to test
    the functions of the module edc_noise_handling.
"""

import numpy as np
import os
import numpy.testing as npt
from pyrato import edc as enh
from numpy import genfromtxt
test_data_path = os.path.join(os.path.dirname(__file__), 'test_data')


def mock_shift_samples_1d(*args, **kwargs):
    return np.array([76])


def mock_shift_samples_2d(*args, **kwargs):
    return np.array([76, 76])


def test_noise_energy_1D():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'noise_energy_1D.csv'),
        delimiter=',')
    actual = enh.estimate_noise_energy(
        rir,
        interval=[0.9, 1.0],
        is_energy=False)
    npt.assert_allclose(actual, expected)


def test_noise_energy_2D():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'noise_energy_2D.csv'),
        delimiter=',')
    actual = enh.estimate_noise_energy(
        rir,
        interval=[0.9, 1.0],
        is_energy=False)
    npt.assert_allclose(actual, expected)


def test_preprocessing_1D():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'preprocessing_1D.csv'),
        delimiter=',')[np.newaxis]
    actual = enh.preprocess_rir(
        rir,
        is_energy=False,
        time_shift=False,
        channel_independent=False)[0]
    npt.assert_allclose(actual, expected)


def test_preprocessing_2D():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'preprocessing_2D.csv'),
        delimiter=',')
    actual = enh.preprocess_rir(
        rir,
        is_energy=False,
        time_shift=False,
        channel_independent=False)[0]
    npt.assert_allclose(actual, expected)


def test_preprocessing_time_shift_1D(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'preprocessing_time_shift_1D.csv'),
        delimiter=',')[np.newaxis]

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_1d)

    actual = enh.preprocess_rir(
        rir,
        is_energy=False,
        time_shift=True,
        channel_independent=False)[0]
    npt.assert_allclose(actual, expected)


def test_preprocessing_time_shift_2D(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'preprocessing_time_shift_2D.csv'),
        delimiter=',')

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_2d)

    actual = enh.preprocess_rir(
        rir,
        is_energy=False,
        time_shift=True,
        channel_independent=False)[0]
    npt.assert_allclose(actual, expected)


def test_preprocessing_time_shift_channel_independent_1D(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(
            test_data_path,
            'preprocessing_time_shift_channel_independent_1D.csv'),
        delimiter=',')[np.newaxis]

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_1d)

    actual = enh.preprocess_rir(
        rir,
        is_energy=False,
        time_shift=True,
        channel_independent=True)[0]
    npt.assert_allclose(actual, expected)


def test_preprocessing_time_shift_channel_independent_2D(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(
            test_data_path,
            'preprocessing_time_shift_channel_independent_2D.csv'),
        delimiter=',')

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_2d)

    actual = enh.preprocess_rir(
        rir,
        is_energy=False,
        time_shift=True,
        channel_independent=True)[0]
    npt.assert_allclose(actual, expected)


def test_smoothed_rir_1D():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'smoothed_rir_1D.csv'),
        delimiter=',')[np.newaxis]
    actual = enh.smooth_rir(
        rir,
        sampling_rate=3000,
        smooth_block_length=0.075)[0]
    npt.assert_allclose(actual, expected)


def test_smoothed_rir_2D():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'smoothed_rir_2D.csv'),
        delimiter=',')
    actual = enh.smooth_rir(
        rir,
        sampling_rate=3000,
        smooth_block_length=0.075)[0]
    npt.assert_allclose(actual, expected)


def test_substracted_1D():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'substracted_1D.csv'),
        delimiter=',')
    actual = enh.subtract_noise_from_squared_rir(rir**2)
    npt.assert_allclose(actual, expected)


def test_substracted_2D():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'substracted_2D.csv'),
        delimiter=',')
    actual = enh.subtract_noise_from_squared_rir(rir**2)
    npt.assert_allclose(actual, expected)


def test_edc_truncation_1D(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'edc_truncation_1D.csv'),
        delimiter=',')

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_1d)

    actual = enh.energy_decay_curve_truncation(
        rir,
        sampling_rate=3000,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True)
    npt.assert_allclose(actual, expected)


def test_edc_truncation_2D(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'edc_truncation_2D.csv'),
        delimiter=',')

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_2d)

    actual = enh.energy_decay_curve_truncation(
        rir,
        sampling_rate=3000,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True)
    npt.assert_allclose(actual, expected)


def test_edc_lundeby_1D(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'edc_lundeby_1D.csv'),
        delimiter=',')

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_1d)

    actual = enh.energy_decay_curve_lundeby(
        rir,
        sampling_rate=3000,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False)
    npt.assert_allclose(actual, expected)


def test_edc_lundeby_2D(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'edc_lundeby_2D.csv'),
        delimiter=',')

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_2d)

    actual = enh.energy_decay_curve_lundeby(
        rir,
        sampling_rate=3000,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False)
    npt.assert_allclose(actual, expected)


def test_edc_lundeby_chu_1D(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'edc_lundeby_chu_1D.csv'),
        delimiter=',')

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_1d)

    actual = enh.energy_decay_curve_chu_lundeby(
        rir,
        sampling_rate=3000,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False)
    npt.assert_allclose(actual, expected)


def test_edc_lundeby_chu_2D(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'edc_lundeby_chu_2D.csv'),
        delimiter=',')

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_2d)

    actual = enh.energy_decay_curve_chu_lundeby(
        rir,
        sampling_rate=3000,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False)
    npt.assert_allclose(actual, expected)


def test_edc_chu_1D(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'edc_chu_1D.csv'),
        delimiter=',')

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_1d)

    actual = enh.energy_decay_curve_chu(
        rir,
        sampling_rate=3000,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False)
    npt.assert_allclose(actual, expected)


def test_edc_chu_2D(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'edc_chu_2D.csv'),
        delimiter=',')

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_2d)

    actual = enh.energy_decay_curve_chu(
        rir,
        sampling_rate=3000,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False)
    npt.assert_allclose(actual, expected)


def test_intersection_time_1D(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'intersection_time_1D.csv'),
        delimiter=',')[np.newaxis].T

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_1d)

    actual = enh.intersection_time_lundeby(
        rir,
        sampling_rate=3000,
        freq='broadband',
        is_energy=False,
        time_shift=False,
        channel_independent=False,
        plot=False)
    npt.assert_allclose(actual, expected)


def test_intersection_time_2D(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'intersection_time_2D.csv'),
        delimiter=',')

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_2d)

    actual = enh.intersection_time_lundeby(
        rir,
        sampling_rate=3000,
        freq='broadband',
        is_energy=False,
        time_shift=False,
        channel_independent=False,
        plot=False)
    npt.assert_allclose(actual, expected)



# -*- coding: utf-8 -*-

"""Top-level package for pyrato."""

__author__ = \
    """Marco Berzborn - Institute for Hearing Technology and Acoustics"""
__email__ = 'marco.berzborn@akustik.rwth-aachen.de'
__version__ = '0.3.1'


from .roomacoustics import (
    reverberation_time_energy_decay_curve,
    schroeder_integration,
    energy_decay_curve_analytic,
    air_attenuation_coefficient,
)
from .dsp import (
    find_impulse_response_maximum,
    find_impulse_response_start,
    filter_fractional_octave_bands,
    time_shift,
    center_frequencies_octaves,
    center_frequencies_third_octaves
)
from .edc import (
    preprocess_rir,
    energy_decay_curve_chu,
    energy_decay_curve_chu_lundeby,
    energy_decay_curve_lundeby,
    energy_decay_curve_truncation,
    estimate_noise_energy,
    intersection_time_lundeby,
)

__all__ = [
    'reverberation_time_energy_decay_curve',
    'schroeder_integration',
    'energy_decay_curve_analytic',
    'air_attenuation_coefficient',
    'find_impulse_response_maximum',
    'find_impulse_response_start',
    'filter_fractional_octave_bands',
    'time_shift',
    'center_frequencies_octaves',
    'center_frequencies_third_octaves',
    'preprocess_rir',
    'energy_decay_curve_chu',
    'energy_decay_curve_chu_lundeby',
    'energy_decay_curve_lundeby',
    'energy_decay_curve_truncation',
    'estimate_noise_energy',
    'intersection_time_lundeby',
]
